#! /usr/bin/python
#
# MSTOPSQL.py Copyright (2008) Gwyneth Morrison
# Convert mssql to postgres
#

import os, re, sys, string

# Symbols for identified source issues

MSNG  = 10		# Missing token
UNSP  = 20		# Unsupported token
SERR  = 30		# Syntax error
INFO  = 40		# Information

trap_stack = [] 	# Trap stack 
debug = 0		# Debug switch
trace = 0		# Trace switch
current_level = 0	# Current level
ansi_warnings = 0	# ansi_warnings 
in_declare = 0		# In declare begin
source_begin = 0	# Start source begin
begin_count = 0		# Generated begin count
begin_stack = []	# Begin stack
level_stack = []	# Level stack
skip_errors = 0		# Skip errors
total_errors = 0	# Total errors
statement_table = {} 	# Statement table
proc_table = {} 	# Proc table
trig_table = {} 	# Trig table
tab_table = {} 		# Table table
output_file = None 	# Secondary file
current_statement = "" 	# Current statement
current_proc = "" 	# Current procedure
trig_proc = "" 		# Trigger procedure
trig_tabname = "" 	# Trigger name
trig_when = "" 		# Trigger when
trig_event = "" 	# Trigger operation
current_column = "" 	# Current column
last_token = None 	# Last printed token 
save_tokens = [] 	# Saved tokens
copy_token = '' 	# Saved copy token
copy_count = 0		# Token copy count
current_line = 1	# Current source line
current_parm = None 	# Current paramater 
current_limit = None 	# Current limit 
return_type = None 	# Return type
intret_table = {} 	# Int Return table
return_default = "void" # Proc return type

#
# Parameter list class
#

parms = []

class parmcase :
  def __init__(self, pname) :
    self.pname = pname
    self.ptype = None
    self.output = None 
    self.default = None
    self.size = None
  def setType(self, ptype) :
    self.ptype = ptype
  def setOutput(self) :
    self.output = "out"
  def setDefault(self, default) :
    self.default = default
  def setSize(self, size) :
    self.size = size
  def printParm(self) :
    if  self.output :
      print "out",self.pname,self.ptype,
    else :
      print "_p%s %s" % (self.pname,self.ptype),
    if self.size : print "(",self.size,")",
  def printDecl(self) :
    if  not self.output :
      print "declare",self.pname,self.ptype,
      if self.size : print "(",self.size,")",
      print "= _p%s" % self.pname,";"
  def listParm(self) :
    if self.ptype == None :
	  self.ptype = "ERROR in TYPE"
    if  self.output :
      outlist = [ "OUT " ] + [ self.ptype ]
    else :
      outlist = [ self.ptype ]
    if self.size : 
      outlist = outlist +  [ "(" ] + [ self.size ] + [ ")" ] 
    return " ".join(outlist)
#
# State token processing routines
#
#
#  Print a warning message
#
def printMessage(token,action,rc) :
  global current_statement
# Replace $$ as it could close a procedure 
  safetoken=re.sub(r'\$\$','XX',token)
  reason = ' MESSAGE '
  if rc == SERR : reason = ' SYNTAX '
  if rc == MSNG : reason = ' MISSING '
  if rc == UNSP : reason = ' UNSUPPORTED '
  if rc == INFO : reason = ' INFORMATION '
  print
  if rc == INFO :
    print "/* INFO in:",current_proc,
  else :
    print "/* WARNING in:",current_proc,
  print "state:",current_level,
  print "line: ",current_line,
  print "in: ",current_statement,
  print safetoken,action,reason," */"
#
#  Print the token
#  Messy business with double operators
#
def printToken(token) :
  global last_token
# Remove ,) errors
  if last_token :
    if last_token == ',' :
      if token == ')' :
        print token
      else : 
        print last_token,token,
      last_token = None
      return
# Fix <> <= >= 
  if last_token :
    if last_token == '<' and token == '>' :
      print "!=",
    else :
      print last_token+token,
    last_token = None
    return
# Save last token
  if token == '<' or token == '>' or token == ',' :
    last_token = token
  if last_token == None : print token,
#
#  Print the stack if debug
#
def printStack(type) :
  global debug
  global current_level
  global level_stack
  if debug :
    print
    print "<",type,":", 
    for l in level_stack :
      print l,
    print "~",current_level,">",
#
#  Return to last state  
#  Preserve the token
#
def skip(token,move,alter) :
  dortn(token,move,alter)
  printStack("Skip")
  if current_level == 349 :
    dlim(token,0,1)
  dotrap(token,move,alter)
  return 2
#
#  Return to last state  
#  Drop token if "OUT"
#  Alter = 1 on keyword
#
def tsto(token,move,alter) :
  global current_level
  global copy_count
  if alter == 1 :
    if return_default == "int" :
      printToken("0 /* inserted */")
    skip(token,move,alter)
    return 2
  if copy_count > 0 :
    dortn(token,move,alter)
    printStack("Tsto")
    dotrap(token,move,alter)
    return 0
  else :
    printToken(token)
    current_level = move
    return 0
#
#  Jump to new state
#  Preserve the token
#
def jump(token,move,alter) :
  global current_level
  current_level = move
  return 2
#
#  Call a new state  
#  Preserve the token
#  Trap the return
#
def trap(token,move,alter) :
  global debug
  global trap_stack
  docall(token,move,alter)
  trap_stack = trap_stack + [ len(level_stack) ]
  if debug :
    print "< Trap", 
    for l in trap_stack :
      print l,
    print ">",
  return 2
#
#  Dump the token
#
def dump(token,move,alter) :
  global current_level
  global output_file
  current_level = move
  if output_file == None :
    output_file = open('output.sql','w')
  output_file.write(token[1:-1])
  output_file.write("go\n")
  print "/* WARNING EXEC procedure dumped */"
  return 0
#
#  Drop the token
#
def drop(token,move,alter) :
  global current_level
  global trace
  current_level = move
  if trace : print "/*^",token,"^*/",
  if alter > 0 : printMessage(token,"dropped",alter)
  return 0
#
#  Set output parameter
#  Drop current token
#
def seto(token,move,alter) :
  global current_level
  global current_parm
  global copy_token
  global copy_count
  current_parm.setOutput()
  copy_count = copy_count + 1
  copy_token =  current_parm.ptype
  current_level = move
  if trace : print "/* seto */",
#
#  Create paramater
#  Drop current token
#
def setp(token,move,alter) :
  global current_level
  global current_parm
  current_parm=parmcase(token)
  current_level = move
  if trace : print "/* setp ",token,"*/",
#
#  Set parameter type
#  Drop current token
#
def sett(token,move,alter) :
  global current_level
  global current_parm
  current_parm.setType(token)
  current_level = move
  if trace : print "/* sett ",token,"*/",
#
#  Set return type
#  Drop current token
#
def setr(token,move,alter) :
  global current_level
  global return_type
  return_type = token
  current_level = move
  if trace : print "/* setr ",token,"*/",
#
#  Set parameter size
#  Drop current token
#
def sets(token,move,alter) :
  global current_level
  global current_parm
  current_parm.setSize(token)
  current_level = move
  if trace : print "/* sets ",token,"*/",
#
#  Save the token 
#
def save(token,move,alter) :
  global current_level
  global current_parm
  global save_tokens
  if trace : print "/* save ",token,"*/",
  save_tokens = save_tokens + [ token ]
  current_level = move
  return 0
#
#  Load last saved token 
#  Preserve current token
#
def lodl(token,move,alter) :
  global current_level
  global save_tokens
  loadlen = len(save_tokens) 
  if loadlen > 0 :
    loadlen = loadlen - 1
    if trace : print "/* lodl ",save_tokens[loadlen],"*/",
    printToken(save_tokens[loadlen])
    del save_tokens[loadlen]
  else :
    print "ERROR load stack empty in ",current_level
  current_level = move
  return 2
#
#  Load first saved token 
#  Preserve current token
#
def load(token,move,alter) :
  global current_level
  global save_tokens
  if len(save_tokens) > 0 :
    if trace : print "/* load ",save_tokens[0],"*/",
    printToken(save_tokens[0])
    del save_tokens[0]
  else :
    print "ERROR load stack empty in ",current_level
  current_level = move
  return 2
#
#  Load all saved tokens 
#  Preserve current token
#
def loda(token,move,alter) :
  global current_level
  global save_tokens
  for stok in save_tokens :
    printToken(stok)
  save_tokens = []
  current_level = move
  return 2
#
#  Drop last saved token 
#  Preserve current token
#
def lodn(token,move,alter) :
  global current_level
  global save_tokens
  loadlen = len(save_tokens) 
  if loadlen > 0 :
    del save_tokens[loadlen - 1]
  current_level = move
  return 2
#
#  Clear saved token stack 
#  Preserve current token
#
def sclr(token,move,alter) :
  global current_level
  global save_tokens
  save_tokens = []
  current_level = move
  return 2
#
#  Save/load the limit 
#  Drop current token
#
def dlim(token,move,alter) :
  global current_level
  global current_limit
  current_level = move
  if alter == 0 :
    if current_limit != None : printMessage("Limit","Reset",SERR)
    current_limit = token
    return 0
  else :
    if current_limit != None :
      printToken("limit")
      printToken(current_limit)
      current_limit = None
    printToken(";")
    return 2
#
#  Print an into list
#  Preserve the token
#
def prto(token,move,alter) :
  global current_level
  global save_tokens
  prcomma = ''
  if len(save_tokens) > 0 :
    printToken("into")
    for stok in save_tokens :
      printToken(prcomma)
      printToken(stok)
      prcomma = ','
    save_tokens = []
  current_level = move
  return 2
#
#  Emit the token unchanged
#
def emit(token,move,alter) :
  global current_level
  printToken(token)
  current_level = move
  return 0
#
#  Set ansi_warnings
#
def ansi(token,move,alter) :
  global current_level
  global ansi_warnings
  current_level = move
  printToken("set escape_string_warning")
  if alter == 0 :
    printToken("= 1")
    ansi_warnings = 1
  else :
    printToken("= 0")
    ansi_warnings = 0
  return 0
#
#  Emit the logged procname 
#
def plog(token,move,alter) :
  global current_level
  printToken(current_proc)
  current_level = move
  return 2
#
#  Emit the trigger header
#
def thed(token,move,alter) :
  global current_proc
  global current_level
  global trig_when
  printToken(current_proc)
  printToken("() returns")
  if trig_when == "instead" :
#    if intret_table.has_key(current_proc.lower()) :
      printToken("int") 
#    else :
#      printToken("void") 
  else:
   printToken("trigger") 
  printToken("as $$ begin") 	
  current_level = move
  return 2
#
#  Emit the logged colname 
#
def cprt(token,move,alter) :
  global current_level
  printToken(current_column)
  current_level = move
  return 2
#
#  Emit the token unchanged
#  Log the procedure name
#
def elog(token,move,alter) :
  global current_level
  global current_proc
  printToken(token)
  current_level = move
  current_proc = token
  return 0
#
#  Emit the procedure name
#
def eprt(token,move,alter) :
  global current_level
  global current_proc
  printToken(current_proc)
  current_level = move
  return 2
#
#  Emit the token unchanged
#  Log the column name
#
def clog(token,move,alter) :
  global current_level
  global current_column
  printToken(token)
  current_level = move
  current_column = token
  return 0
#
#  Emit the token unchanged
#  Log the table name
#
def ctab(token,move,alter) :
  global current_level
  printToken(token)
  current_level = move
  tab_table[token.lower()] = "CREATE"
  return 0
#
#  Emit the token unchanged
#  Log the table drop
#
def dtab(token,move,alter) :
  global current_level
  printToken(token)
  current_level = move
  tab_table[token.lower()] = "DROP"
  return 0
#
#  Log trigger parameters
#
def tlog(token,move,alter) :
  global current_level
  global current_proc
  global trig_tabname
  global trig_proc
  global trig_table
  global trig_when
  global trig_event
  if alter == 0 :
    current_proc = token
  if alter == 1 :
    trig_table[current_proc.lower()] = token
    trig_tabname = token
    trig_proc = current_proc
  if alter == 2 :
    trig_when = token.lower()
    if trig_when == "for" :
      trig_when = "before"
  if alter == 3 :
    trig_event = token
  current_level = move
  return 0
#
#  Print the trigger table
#
def tprt(token,move,alter) :
  global current_level
  global current_proc
  global trig_table
  current_level = move
  if trig_table.has_key(current_proc.lower()) :
    printToken(trig_table[current_proc.lower()])
  else :
    printToken("sysobjects")
    print "/* WARNING trigger proc",current_proc,"missing */"
  return 2
#
#  Strip and emit  the token 
#
def strp(token,move,alter) :
  global current_level
  newtoken=re.sub(r'\'','"',token)
#  newtoken=re.sub(r'\ ','_',newtoken)
#  newtoken=re.sub(r'%','',newtoken)
#  newtoken=re.sub(r'\/','',newtoken)
  printToken(newtoken)
  current_level = move
  return 0
#
#  Lower and emit the token 
#
def lowr(token,move,alter) :
  global current_level
  printToken(token.lower())
  current_level = move
  return 0
#
#  Escape quote emit the token 
#
def escp(token,move,alter) :
  global current_level
  if token [ 0 ] == 'b' and token [ 1 ] == '\'' :
    printToken(token)
    return 0
  char_range = range(len(token))
  newtoken = ""
  for i in char_range :
    newtoken = newtoken + token [ i ]  
    if token [ i ] == '\'' and i > 0 and i < len(token) - 1 :
      newtoken = newtoken + '\''
  printToken(newtoken)
  current_level = move
  return 0
#
#  Insert the new token 
#  Preserve the current token
#  Close a declare begin block
#
def isrb(token,move,alter) :
  global begin_count
  global in_declare
  isrt(token,move,alter)
  in_declare = 0
  begin_count = begin_count + 1
  return 2
#
#  Add a paramater to the list
#  Preserve the current token
#
def addp(token,move,alter) :
  global current_level
  global current_parm
  global parms
  parms = parms + [ current_parm ]
  current_parm = None 	
  current_level = move
  return 2
#
#  Insert parameter list
#  Preserve the current token
#
def isrp(token,move,alter) :
  global intret_table
  global current_level
  global copy_count
  global current_parm
  global return_default
  global current_proc
  global return_type
  global parms
  if current_parm :
    addp(token,move,alter)
  print "("
  comma = ""
  parm_list = [ "(" ]
  for parm in parms :
    print comma
    parm.printParm() 
    parm_list = parm_list + [ comma ]
    parm_list = parm_list + [ parm.listParm() ]
    comma = ","
  print ")"
  parm_list = parm_list + [ ")" ]
  proc_table[current_proc.lower()] = " ".join(parm_list)
  if intret_table.has_key(current_proc.lower()) :
    return_default = "int"
  else :
    return_default = "void"
  print "returns",
  if return_type : 
    printToken(return_type)
    return_type = None
  else :
    if copy_count == 0 : printToken(return_default)
    if copy_count > 1 : printToken("record")
    if copy_count == 1 : printToken(copy_token)
  print "as $$"
  for parm in parms :
    parm.printDecl()
  print "begin"
  parms = []
  current_level = move
  return 2
#
#  Insert current_proc signature
#  Preserve token
#  Clear current proc if alter
#
def isig(token,move,alter) :
  global current_level
  global current_proc
  global trace
  if proc_table.has_key(current_proc.lower()) :
    printToken(proc_table[current_proc.lower()])
  else:
    print "()"
  current_level = move
  if alter > 0 : current_proc = ""
  return 2
#
#  Insert the new token 
#  Preserve the current token
#
def isrt(token,move,alter) :
  global current_level
  global trace
  safetoken=re.sub(r'\$\$','XX',token)
  if trace : print "/*>",safetoken,"<*/",
  if alter > 0 : printMessage(token,"inserted",alter)
  printToken(token)
  current_level = move
  return 2
#
#  Execute a return only
#
def dortn(token,move,alter) :
  global level_stack
  global current_level
  last = len(level_stack) - 1
  if last >= 0:
    current_level = level_stack[last]
    del level_stack[last]
  else : current_level = 0
  if current_level == 0 : printToken(";")
#
#  Attempt to trap
#
def dotrap(token,move,alter) :
  global level_stack
  global current_level
  global trap_stack
  last = len(trap_stack) -1
  if last >= 0 and len(level_stack) == trap_stack[last] :
    del trap_stack[last]
    last = len(level_stack) - 1
    if last >= 0:
      current_level = level_stack[last]
      del level_stack[last]
    else : current_level = 0
    if current_level == 0 : printToken(";")
    printStack("Dotrap")
#
#  Emit token and return
#
def rtne(token,move,alter) :
  printToken(token)
  dortn(token,move,alter)
  printStack("Ret")
  dotrap(token,move,alter)
  return 0
#
#  Drop token and return
#
def rtnd(token,move,alter) :
  global trace
  dortn(token,move,alter)
  printStack("Ret")
  if trace : print "/*^",token,"^*/",
  if alter > 0 : printMessage(token,"dropped",alter) 
  return 0
#
#  Preserve token and move
#  Dump ends to close begins
#
def endb(token,move,alter) :
  global begin_count
  global begin_stack
  global current_level
  current_level = move
  if trace : print "/* Endblock */"
  printToken(";")
  for i in range(begin_count) :
    printToken("end;")
  last = len(begin_stack) - 1
  if last >= 0:
    begin_count = begin_stack[last]
    del begin_stack[last]
  else : 
    begin_count = 0
  return 2
#
#  Preserve token and return
#  Closes a procedure
#  Flushes the stack
#
def endp(token,move,alter) :
  global level_stack
  global current_level
  global begin_stack
  global begin_count
  global copy_count
  global proc_type
  global current_proc
  printToken("null")
  endb(token,move,alter)
  begin_count = 0
  begin_stack = []
  proc_type = None
  copy_count = 0
  if current_proc == "" : return
  dortn(token,move,alter)
  while current_level != 282 and current_level != 298 and len(level_stack) > 0:
    dortn(token,move,alter)
    printStack("Ret")
    dotrap(token,move,alter)
  if trace : print "/* Endproc */"
  current_proc = ""
  return 2
#
#  Preserve token and move
#  Generates a procedure
#
def endt(token,move,alter) :
  global current_level
  global trig_proc
  dortn(token,move,alter)
  if trig_proc != "" :
    if trig_when == "instead":
      printToken("create rule")
      printToken(trig_proc)
      printToken("as on")
      printToken(trig_event)
      printToken("to")
      printToken(trig_tabname)
      printToken("do instead")
      printToken("select")
      printToken(trig_proc)
      printToken("();")
    else:
      printToken("create trigger")
      printToken(trig_proc)
      printToken(trig_when)
      printToken(trig_event)
      printToken("on")
      printToken(trig_tabname)
      printToken("execute procedure")
      printToken(trig_proc)
      printToken("();")
    trig_proc = ""
  return 2
#
#  Execute a call only
#
def docall(token,move,alter) :
  global level_stack
  global current_level
  if alter > 0 :
    level_stack = level_stack + [ alter ]
  else :
    level_stack = level_stack + [ current_level ]
  current_level = move
  printStack("Call")
#
#  Emit token and call
#
def call(token,move,alter) :
  printToken(token)
  docall(token,move,alter)
  return 0
#
#  Preserve token and call
#
def calp(token,move,alter) :
  docall(token,move,alter)
  return 2
#
#  Start a source begin block
#  Stack the current begin count
#  Preserve token and call
#
def stab(token,move,alter) :
  global begin_stack
  global begin_count
  begin_stack = begin_stack + [ begin_count ]
  begin_count = 0
  docall(token,move,alter)
  return 2
#
#  Drop token and call
#
def cald(token,move,alter) :
  docall(token,move,alter)
  if trace : print "/*^",token,"^*/",
#  printMessage(token,"dropped",0);
  return 0
#
#  Replace token and call
#
def calr(token,move,alter) :
  dorepl(token,move,0)
  docall(token,move,alter)
  if trace : print "/*^",token,"^*/",
#  printMessage(token,"replaced",0);
  return 0
#
#  Declare begin block call
#
def calb(token,move,alter) :
  global in_declare
  printToken(token)
  docall(token,move,alter)
  in_declare = 1
  return 0

#
# Replacement case table for MSSQL 
# state number, origional token, new token
# A state number of zero means replace always
#

repls = []

class rcase :
  def __init__(self, level, token, alter) :
    self.level = level
    self.token = token
    self.alter = alter

def addRepl(level, token, alter) :
  global repls
  repls = repls + [ rcase(level, token, alter) ]

def addAllRepls() :

  addRepl(284,	"as",			";") 	
  addRepl(304,	"s",			"'s'") 	
  addRepl(304,	"hour",			"'hour'") 	
  addRepl(304,	"minute",		"'minute'") 	
  addRepl(323,	"name",			"id") 	
  addRepl(500,	"trigger",		"trigger if exists") 	
  addRepl(810,	",",			"; declare") 	
  addRepl(1150,	"1",			"=1") 	
  addRepl(1150,	"0",			"=0") 	
  addRepl(1206,	"0",			"found") 	
  addRepl(1310,	"go",			"()") 	
  addRepl(1402,	"exec", 		"select") 	
  addRepl(1404,	"exec", 		"select") 	
  addRepl(1800,	"procedure",		"drop function") 	
  addRepl(1800,	"function",		"drop function") 	
  addRepl(1800,	"table",		"alter table") 	
  addRepl(1800,	"database",		"alter database") 	
  addRepl(1800,	"trigger",		"drop trigger if exists") 	
  addRepl(1920,	",",			"using") 	
  addRepl(1625,	"on",			"where") 	
  addRepl(2515,	"on",			"where") 	
  addRepl(0,	"where",		"and") 	
  addRepl(0,	"to",			"tot") 	
  addRepl(0,	"char",			"chr") 	
  addRepl(0,	"from",			"using") 	
  addRepl(0,	"table",		"setof record") 	
  addRepl(0,	"constraint",		"owner to postgres") 	
  addRepl(0,	"raiserror",		"raise exception '%s %d %d',") 	
  addRepl(0,	"db_name",		"current_database") 	
  addRepl(0,	"procedure",		"function") 	
  addRepl(0,	"uniqueidentifier",	"uuid") 	
  addRepl(0,	"nvarchar",		"varchar") 	
  addRepl(0,	"local",		"no hold") 	
  addRepl(0,	"forward_only",		"no scroll") 	
  addRepl(0,	"sysname",		"varchar(128)") 	
  addRepl(0,	"ntext",		"text") 	
  addRepl(0,	"bit",			"boolean") 	
  addRepl(0,	"tinyint",		"smallint") 	
  addRepl(0,	"varbinary",		"bit varying") 	
  addRepl(0,	"money",		"numeric(19,4)") 	
  addRepl(0,	"sql_variant",		"varchar") 	
  addRepl(0,	"user",			"\"user\"") 	
  addRepl(0,	"offset",		"\"offset\"") 	
  addRepl(0,	"datetime",		"timestamp") 	
  addRepl(0,	"identity",		"default identity") 	
  addRepl(0,	"on",			"= 1") 	
  addRepl(0,	"off",			"= 0") 	
  addRepl(0,	"mdy",			"to iso,mdy") 	
  addRepl(0,	"global",		"= 0") 	
  addRepl(0,	"simple",		"= 0") 	
  addRepl(0,	"exec", 		"perform") 	
  addRepl(0,	"=",			":=") 	
  addRepl(0,	"transaction",		"null;") 	
  addRepl(0,	"tran",			"null;") 	
  addRepl(0,	"rollback",		"raise exception 'ROLLBACK';") 	
  addRepl(0,	"deallocate",		"null") 	
  addRepl(0,	"print",		"raise info '%s',") 	
  addRepl(0,	"break",		"exit;") 	
  addRepl(0,	"read_write",		"owner to postgres") 	
  addRepl(0,	"multi_user",		"set owner to postgres") 	
  addRepl(0,	"0x0",			"null") 	
  addRepl(0,	"'8.0.7'",		"0") 	
#
# Execute a replace only
#
def dorepl(token,move,alter) :
  global current_level
  global repls
  global trace
  for repl in repls :
    if token.lower() == repl.token :
      if  current_level == repl.level or repl.level == 0 : 
        if alter > 0 : printMessage(token,"replaced",alter) 
  	if trace : print "/*=",token,"=*/",
        printToken(repl.alter)
        return
  print "/* ERROR replacement failed for", token, "*/",
#
#  Replace token and call
#
def repc(token,move,alter) :
  dorepl(token,move,0)
  docall(token,move,alter)
#
#  Replace token and return
#
def repr(token,move,alter) :
  dorepl(token,move,alter)
  dortn(token,move,alter)
  printStack("Ret")
  dotrap(token,move,alter)
#
#  Replace token and move
#
def repl(token,move,alter) :
  global current_level
  dorepl(token,move,alter)
  current_level = move
#
#  Replace token and save 
#  Only used for paramaters
#
def reps(token,move,alter) :
  global repls
  global trace
  global current_level
  global current_parm
  global save_tokens
  for repl in repls :
    if token.lower() == repl.token :
      if  current_level == repl.level or repl.level == 0 : 
        if alter > 0 : printMessage(token,"replaced",alter) 
        if trace : print "/* reps ",repl.alter,"*/",
        save_tokens = save_tokens + [ repl.alter ]
        current_level = move
        return
  print "/* ERROR replacement failed for", token, "*/",
  current_level = move
#
#  Replace token set type
#  Only used for paramaters
#
def rept(token,move,alter) :
  global repls
  global trace
  global current_level
  global current_parm
  for repl in repls :
    if token.lower() == repl.token :
      if  current_level == repl.level or repl.level == 0 : 
        if alter > 0 : printMessage(token,"replaced",alter) 
        if trace : print "/* rept ",repl.alter,"*/",
        current_parm.setType(repl.alter)
        current_level = move
        return
  print "/* ERROR replacement failed for", token, "*/",
  current_level = move
#
#  Replace token and set return type
#  Only used for return types
#
def retr(token,move,alter) :
  global repls
  global trace
  global current_level
  global return_type
  for repl in repls :
    if token.lower() == repl.token :
      if  current_level == repl.level or repl.level == 0 : 
        if alter > 0 : printMessage(token,"replaced",alter) 
        if trace : print "/* retr ",repl.alter,"*/",
        return_type = repl.alter
        current_level = move
        return
  print "/* ERROR replacement failed for", token, "*/",
  current_level = move
#
# Parse case table for SQL		
#
class parsecase :
  def __init__(self, level, action, move, alter, token) :
    self.level = level
    self.action = action
    self.move = move
    self.alter = alter
    self.token = token

cases = []

#
#  Add a case to the case table
#
def addCase(level, action, move, alter, token) :
  global cases
  cases = cases + [ parsecase(level, action, move, alter, token) ]
#
#  Print the case table
#
def printCases() :
  global cases
  actions = {}
  for case in cases :
    actions[getattr(case.action,"__name__")] = 0
  for key in actions :
    print "def",key,"(token, move, alter) : "
    print "  return \"%s\"" % key
    print "#"
  print
  print "def addcases() :"
  print
  for case in cases :
    print "  addCase(%s,\t%s," % (case.level,getattr(case.action,"__name__")),
    print "\t%s,\t%s,\t\"%s\")" % (case.move,case.alter,case.token)

def addAllCases() :

# All statements 

  addCase(0,	drop,	1100,	0,	"identity_insert") # delete? 
  addCase(0,	drop,	1140,	INFO,	"ansi_padding") # delete? 
  addCase(0,	drop,	0,	0,	"textimage_on") # delete? 
  addCase(0,	drop,	0,	0,	"on") 		# delete? 
  addCase(0,	drop,	0,	0,	"primary") 	# delete? 

  addCase(0,	call,	100,	0,	"create") 	# create 
  addCase(0,	call,	300,	349,	"select") 	# select
  addCase(0,	repc,	400,	0,	"raiserror") 	# raiserror
  addCase(0,	call,	500,	0,	"drop") 	# drop
  addCase(0,	call,	550,	0,	"truncate") 	# truncate
  addCase(0,	calr,	600,	0,	"print") 	# print
  addCase(0,	call,	750,	0,	"if") 		# if
  addCase(0,	emit,	0,	0,	"else") 	# else
  addCase(0,	repl,	0,	0,	"break") 	# break
  addCase(0,	jump,	1750,	0,	"begin") 	# begin
  addCase(0,	jump,	1775,	0,	"end") 		# end
  addCase(0,	drop,	0,	UNSP,	"commit") 	# commit
  addCase(0,	repl,	0,	0,	"rollback") 	# rollback
  addCase(0,	drop,	0,	0,	"transaction") 	# transaction
  addCase(0,	drop,	0,	0,	"tran") 	# tran
  addCase(0,	endp,	0,	0,	"go") 		# go
  addCase(0,	drop,	0,	0,	";") 		# ;
  addCase(0,	calb,	800,	0,	"declare") 	# declare
  addCase(0,	call,	900,	0,	"open") 	# open
  addCase(0,	call,	900,	0,	"close") 	# close
  addCase(0,	calr,	910,	0,	"deallocate") 	# deallocate
  addCase(0,	call,	1000,	0,	"fetch") 	# fetch
  addCase(0,	cald,	1100,	0,	"set") 		# set
  addCase(0,	call,	1250,	0,	"while") 	# while
  addCase(0,	cald,	1300,	0,	"exec") 	# exec
  addCase(0,	cald,	1300,	0,	"execute") 	# exeutec
  addCase(0,	call,	1400,	0,	"insert") 	# insert
  addCase(0,	call,	1500,	0,	"return") 	# return
  addCase(0,	call,	1600,	0,	"update") 	# update
  addCase(0,	cald,	1700,	0,	"use") 		# use
  addCase(0,	cald,	1800,	0,	"alter") 	# alter
  addCase(0,	call,	1900,	0,	"delete") 	# delete

# Create 

  addCase(100,	emit,	110,	0,	"table") 	# create table
  addCase(100,	emit,	140,	0,	"view") 	# create view
  addCase(100,	repl,	240,	0,	"procedure") 	# function
  addCase(100,	emit,	240,	0,	"function") 	# function
  addCase(100,	drop,	290,	0,	"trigger") 	# trigger
  addCase(100,	drop,	100,	INFO,	"clustered") 	
  addCase(100,	drop,	100,	0,	"nonclustered") 	
  addCase(100,	emit,	100,	0,	"unique") 	
  addCase(100,	emit,	100,	0,	"or") 	
  addCase(100,	emit,	100,	0,	"replace") 	
  addCase(100,	emit,	230,	0,	"index") 	# create index

  addCase(110,	emit,	130,	0,	"(") 		# missing name
  addCase(110,	ctab,	115,	0,	"<value>") 	# table name 
  addCase(115,	emit,	130,	0,	"(") 	
  addCase(115,	emit,	110,	0,	".") 		# schema.table

  addCase(130,	emit,	210,	0,	")") 	
  addCase(130,	emit,	170,	0,	"constraint") 	
  addCase(130,	emit,	210,	0,	"primary") 	
  addCase(130,	drop,	130,	0,	"asc") 		
  addCase(130,	repl,	132,	0,	"user") 	# column name
  addCase(130,	repl,	132,	0,	"offset") 	# column name
  addCase(130,	emit,	132,	0,	"<value>") 	# column name

  addCase(132,	repl,	134,	0,	"tinyint") 	# column type
  addCase(132,	repl,	134,	0,	"datetime") 	# column type
  addCase(132,	repl,	134,	0,	"nvarchar") 	# column type
  addCase(132,	repl,	134,	0,	"varbinary") 	# column type
  addCase(132,	repl,	134,	0,	"money") 	# column type
  addCase(132,	repl,	134,	0,	"bit") 		# column type
  addCase(132,	repl,	134,	0,	"ntext") 	# column type
  addCase(132,	repl,	134,	0,	"uniqueidentifier") 
  addCase(132,	emit,	134,	0,	"<value>") 	# column type

  addCase(134,	emit,	130,	0,	",") 	
  addCase(134,	call,	136,	0,	"(") 	
  addCase(134,	rtne,	0,	0,	")") 	
  addCase(134,	emit,	134,	0,	"not") 	
  addCase(134,	emit,	134,	0,	"null") 	
  addCase(134,	drop,	134,	INFO,	"rowguidcol") 	
  addCase(134,	repl,	134,	0,	"identity") 	
  addCase(134,	drop,	139,	0,	"collate") 	
  addCase(134,	cald,	150,	0,	"constraint") 	
  addCase(134,	drop,	135,	0,	"on") 	
  addCase(134,	emit,	210,	0,	"primary") 	
  addCase(134,	call,	152,	0,	"default") 	
  addCase(134,	skip,	0,	0,	"<any>") 	

  addCase(135,	drop,	134,	0,	"primary") 	

  addCase(136,	call,	300,	138,	"select") 	 
  addCase(136,	emit,	138,	0,	"<value>") 	# column size

  addCase(138,	rtne,	0,	0,	")") 	
  addCase(138,	call,	138,	0,	"(") 		# function
  addCase(138,	emit,	136,	0,	",") 	
  addCase(138,	skip,	0,	0,	"<any>") 	

  addCase(139,	drop,	134,	0,	"<value>") 	# collate

  addCase(140,	emit,	142,	0,	"<value>") 	# view name 
  addCase(142,	drop,	142,	0,	"with") 	 
  addCase(142,	drop,	142,	0,	"schemabinding") 	 
  addCase(142,	emit,	144,	0,	"as") 	 
  addCase(142,	emit,	146,	0,	"(") 	 
  addCase(142,	emit,	140,	0,	".") 	 
  addCase(142,	skip,	0,	0,	"<any>") 	 

  addCase(144,	call,	300,	0,	"select") 	 
  addCase(144,	emit,	144,	0,	"union") 	 
  addCase(144,	emit,	140,	0,	"as") 	 
  addCase(144,	skip,	0,	0,	"<any>") 	 

  addCase(146,	emit,	148,	0,	"<value>") 	# view colname 
  addCase(148,	emit,	146,	0,	",") 	
  addCase(148,	emit,	144,	0,	")") 	

  addCase(150,	save,	151,	0,	"<value>") 	# constraint name
  addCase(151,	emit,	153,	0,	"default") 	
  addCase(151,	save,	157,	0,	"check") 	
  addCase(151,	save,	160,	0,	"primary") 	
  addCase(151,	save,	160,	0,	"unique") 	
  addCase(151,	skip,	0,	0,	"<any>") 	 

  addCase(152,	call,	152,	0,	"(") 		# default
  addCase(152,	rtne,	0,	0,	")") 	
  addCase(152,	emit,	152,	0,	"-") 	
  addCase(152,	skip,	0,	0,	",") 	
  addCase(152,	drop,	150,	0,	"constraint") 	
  addCase(152,	repl,	154,	INFO,	"'8.0.7'") 	
  addCase(152,	emit,	154,	0,	"<value>") 	

  addCase(153,	sclr,	152,	0,	"<any>") 	

  addCase(154,	call,	155,	0,	"(") 		# function
  addCase(154,	rtne,	0,	0,	")") 	
  addCase(154,	emit,	152,	0,	".") 	
  addCase(154,	skip,	0,	0,	"<any>") 	
  
  addCase(155,	rtne,	0,	0,	")") 	
  addCase(155,	emit,	156,	0,	"<value>") 	

  addCase(156,	emit,	155,	0,	",") 	
  addCase(156,	rtne,	0,	0,	")") 	

  addCase(157,	isrt,	158,	0,	"constraint") 	
  addCase(158,	loda,	159,	0,	"<any>") 	

  addCase(159,	call,	159,	134,	"(") 		# check
  addCase(159,	rtne,	0,	0,	")") 	
  addCase(159,	emit,	159,	0,	"<value>") 	

  addCase(160,	save,	162,	0,	"key") 		
  addCase(160,	drop,	160,	0,	"nonclustered") 	
  addCase(160,	drop,	160,	0,	"clustered") 	

  addCase(162,	drop,	162,	0,	"nonclustered") 	
  addCase(162,	drop,	162,	0,	"clustered") 	
  addCase(162,	drop,	164,	0,	"(") 		# table constraint
  addCase(162,	isrt,	163,	0,	"constraint") 	
  addCase(163,	loda,	210,	0,	"<any>") 	

  addCase(164,	isrt,	165,	0,	", constraint") 	
  addCase(165,	loda,	166,	0,	"<any>") 	
  addCase(166,	isrt,	167,	0,	"(") 	
  addCase(167,	call,	130,	210,	"<value>") 	

  addCase(170,	emit,	171,	0,	"<value>") 	# constraint name
  addCase(171,	emit,	159,	0,	"check") 	
  addCase(171,	emit,	210,	0,	"primary") 	
  addCase(171,	emit,	210,	0,	"unique") 	
  addCase(171,	skip,	0,	0,	"<any>") 	 

  addCase(210,	emit,	210,	0,	"key") 		# primary key
  addCase(210,	drop,	210,	0,	"clustered") 	
  addCase(210,	drop,	210,	0,	"nonclustered") 	
  addCase(210,	emit,	210,	0,	"unique") 	
  addCase(210,	call,	220,	0,	"(") 	
  addCase(210,	drop,	210,	0,	"on") 	
  addCase(210,	drop,	210,	0,	"textimage_on") 	
  addCase(210,	drop,	210,	0,	"primary") 	
  addCase(210,	emit,	130,	0,	",") 	
  addCase(210,	rtne,	0,	0,	")") 	
  addCase(210,	skip,	0,	0,	"<any>") 	

  addCase(220,	emit,	222,	0,	"<value>") 	
  addCase(222,	drop,	222,	0,	"asc") 		
  addCase(222,	rtne,	0,	0,	")") 	
  addCase(222,	emit,	220,	0,	",") 	

  addCase(230,	emit,	232,	0,	"on") 		# index
  addCase(230,	emit,	230,	0,	"<value>") 

  addCase(232,	emit,	234,	0,	"(") 
  addCase(232,	emit,	232,	0,	".") 
  addCase(232,	emit,	232,	0,	"<value>") 

  addCase(234,	emit,	234,	0,	",") 
  addCase(234,	emit,	236,	0,	")") 
  addCase(234,	emit,	234,	0,	"<value>") 

  addCase(236,	drop,	236,	0,	"on") 
  addCase(236,	drop,	236,	0,	"primary") 
  addCase(236,	skip,	0,	0,	"<any>") 	
 
# Function
 
  addCase(240,	elog,	242,	0,	"<value>") 	# procedure name

  addCase(242,	drop,	260,	0,	"as") 		# no parameters
  addCase(242,	drop,	242,	0,	"(") 
  addCase(242,	drop,	260,	0,	")") 		
  addCase(242,	setp,	243,	0,	"<value>") 	# parameter name

  addCase(243,	rept,	250,	0,	"nvarchar") 	# parameter type
  addCase(243,	rept,	250,	0,	"uniqueidentifier")
  addCase(243,	rept,	250,	0,	"datetime")
  addCase(243,	rept,	250,	0,	"sql_variant")
  addCase(243,	rept,	250,	0,	"varbinary")
  addCase(243,	rept,	250,	0,	"ntext")
  addCase(243,	rept,	250,	0,	"tinyint")
  addCase(243,	rept,	250,	0,	"bit")
  addCase(243,	rept,	250,	UNSP,	"sysname")
  addCase(243,	drop,	243,	0,	"as")		# is valid in MS
  addCase(243,	drop,	245,	0,	"returns") 
  addCase(243,	sett,	250,	0,	"<value>") 	# type

  addCase(245,	call,	110,	0,	"table") 
  addCase(245,	drop,	260,	0,	"as") 		
  addCase(245,	skip,	0,	0,	"<key>") 
  addCase(245,	setr,	245,	0,	"<value>") 	# return type

  addCase(250,	drop,	270,	0,	"(") 		# size
  addCase(250,	drop,	253,	0,	"=") 		# default
  addCase(250,	seto,	250,	0,	"output") 	# output parameter	
  addCase(250,	drop,	260,	0,	"as") 		
  addCase(250,	drop,	256,	0,	")") 		
  addCase(250,	addp,	255,	0,	"<any>") 	
  addCase(253,	drop,	253,	0,	"-") 	
  addCase(253,	drop,	250,	0,	"<value>") 	

  addCase(255,	drop,	242,	0,	",") 
  addCase(255,	emit,	245,	0,	"returns") 
  addCase(255,	emit,	256,	0,	")") 

  addCase(256,	drop,	280,	0,	"returns") 
  addCase(256,	drop,	260,	0,	"as") 		

  addCase(260,	drop,	280,	0,	"returns") 	# returns given
  addCase(260,	isrp,	261,	0,	"<any>") 	# insert parameter list
  addCase(261,	calp,	0,	282,	"<any>") 	# procedure definition

  addCase(270,	sets,	273,	0,	"<value>") 	# size given

  addCase(273,  seto,	273,	0,	"output")
  addCase(273,  drop,	278,	0,	"=")
  addCase(273,  drop,	273,	0,	")") 	
  addCase(273,	drop,	260,	0,	"as") 		
  addCase(273,	drop,	280,	0,	"returns") 		
  addCase(273,	addp,	255,	0,	"<any>") 	# add paramater 

  addCase(278,  drop,	273,	0,	"<value>")

  addCase(280,	retr,	281,	0,	"table") 	# returns
  addCase(280,	setr,	281,	0,	"int") 
  addCase(280,	retr,	281,	0,	"bit") 
  addCase(280,	retr,	281,	0,	"datetime") 
  addCase(280,	setr,	281,	0,	"float") 
  addCase(280,	retr,	281,	0,	"nvarchar") 
  addCase(280,	setr,	281,	0,	"varchar") 
  addCase(280,	setr,	281,	0,	"bigint") 
  addCase(280,	retr,	281,	0,	"uniqueidentifier") 
  addCase(280,	drop,	260,	0,	"as") 		
  addCase(280,	setr,	284,	0,	"<value>") 	# table name
  
  addCase(281,	drop,	281,	0,	"(") 
  addCase(281,	drop,	281,	0,	")") 
  addCase(281,	drop,	281,	0,	"with") 
  addCase(281,	drop,	281,	0,	"schemabinding") 
  addCase(281,	drop,	260,	0,	"as") 
  addCase(281,	drop,	281,	0,	"<value>") 

  addCase(282,	isrt,	283,	0,	"end; $$ language plpgsql") 	
  addCase(283,	endt,	0,	0,	"<any>") 

  addCase(284,	retr,	285,	0,	"table") 
  addCase(284,	repl,	284,	0,	"as") 
  addCase(284,	drop,	284,	0,	"with") 
  addCase(284,	drop,	284,	0,	"schemabinding") 
  addCase(284,	calp,	0,	282,	"<any>") 

  addCase(285,	isrp,	286,	0,	"<any>") 
  addCase(286,	isrt,	287,	0,	"create temp table bogus") 

  addCase(287,	call,	130,	0,	"(") 
  addCase(287,	drop,	287,	0,	"with") 	
  addCase(287,	drop,	287,	0,	"schemabinding") 	
  addCase(287,	drop,	287,	0,	"as") 	
  addCase(287,	isrt,	284,	0,	";") 	

  # Trigger

  addCase(290,	drop,	291,	0,	"on")
  addCase(290,	tlog,	290,	0,	"<value>") 	# trigger name

  addCase(291,	tlog,	292,	1,	"<value>")	# table name

  addCase(292,	drop,	293,	0,	"as")		# on 
  addCase(292,	tlog,	292,	2,	"for")
  addCase(292,	drop,	292,	0,	"of")
  addCase(292,	tlog,	292,	2,	"instead")
  addCase(292,	tlog,	292,	2,	"after")
  addCase(292,	tlog,	292,	2,	"before")
  addCase(292,	tlog,	292,	3,	"delete")
  addCase(292,	tlog,	292,	3,	"insert")
  addCase(292,	tlog,	292,	3,	"update")
  addCase(292,	drop,	292,	0,	",")
  addCase(292,	drop,	292,	0,	"<value>")

  addCase(293,	isrt,	294,	0,	"function")
  addCase(294,	thed,	296,	0,	"<any>")
  addCase(296,	calp,	0,	298,	"<any>")

  addCase(298,	isrt,	299,	0,	"return null; end; $$ language plpgsql") 	
  addCase(299,	endt,	0,	0,	"<any>") 

# Select 

  addCase(300,	emit,	310,	0,	"*") 
  addCase(300,	repl,	310,	0,	"db_name")
  addCase(300,	drop,	308,	0,	"top") 
  addCase(300,	call,	300,	310,	"(") 		
  addCase(300,	emit,	300,	0,	"-") 		
  addCase(300,	emit,	300,	0,	",") 		
  addCase(300,	call,	360,	310,	"case") 		
  addCase(300,	call,	300,	0,	"select") 		
  addCase(300,	emit,	300,	0,	"distinct") 		
  addCase(300,	emit,	303,	0,	"dateadd") 		
  addCase(300,	emit,	303,	0,	"datepart") 		
  addCase(300,	emit,	311,	0,	"as") 	
  addCase(300,	emit,	301,	0,	"into") 	
  addCase(300,	prto,	320,	0,	"from") 	
  addCase(300,	cald,	1110,	310,	"convert") 	
  addCase(300,	skip,	0,	0,	"<key>") 	
  addCase(300,	save,	302,	0,	"<symbol>") 	# symbol 
  addCase(300,	reps,	306,	0,	"user") 	# colname 
  addCase(300,	reps,	306,	0,	"offset") 	# colname 
  addCase(300,	save,	306,	0,	"<value>") 	# colname 

  addCase(301,	emit,	300,	0,	"<value>") 	

  addCase(302,	drop,	300,	0,	"=") 		# assignment
  addCase(302,	load,	310,	0,	"<any>") 	

  addCase(303,	call,	304,	310,	"(") 		# dateadd
  addCase(304,	repl,	300,	0,	"s") 	
  addCase(304,	repl,	300,	0,	"hour") 	
  addCase(304,	repl,	300,	0,	"minute") 	
  addCase(304,	emit,	300,	0,	"<any>") 	

  addCase(306,	drop,	307,	UNSP,	"=") 		# column alias
  addCase(306,	lodl,	310,	0,	"<any>") 	
  addCase(307,	lodn,	300,	0,	"<any>") 	

  addCase(308,	dlim,	309,	0,	"<value>") 	# top limit	
  addCase(309,	drop,	300,	0,	"percent") 	
  addCase(309,	jump,	300,	0,	"<any>") 	

  addCase(310,	prto,	320,	0,	"from") 	# columns
  addCase(310,	emit,	300,	0,	",") 
  addCase(310,	emit,	300,	0,	"into") 	# columns
  addCase(310,	call,	315,	0,	"(") 		# function
  addCase(310,	rtne,	0,	0,	")") 		# end expr
  addCase(310,	emit,	300,	0,	"+") 
  addCase(310,	emit,	300,	0,	"-") 
  addCase(310,	emit,	300,	0,	"/") 
  addCase(310,	emit,	300,	0,	"?") 
  addCase(310,	emit,	300,	0,	"*") 
  addCase(310,	emit,	300,	0,	"=") 		# assignment
  addCase(310,	emit,	300,	0,	".") 	
  addCase(310,	emit,	338,	0,	"order") 	
  addCase(310,	drop,	310,	UNSP,	"collate") 	
  addCase(310,	emit,	311,	0,	"as") 	
  addCase(310,	prto,	313,	0,	"<key>") 	
  addCase(310,	isrt,	311,	0,	"as") 	

  addCase(311,	strp,	312,	0,	"<value>") 	# name
  addCase(312,	emit,	355,	0,	"union")
  addCase(312,	emit,	300,	0,	",") 	
  addCase(312,	prto,	320,	0,	"from")
  addCase(312,	emit,	300,	0,	"into") 	# columns
  addCase(312,	call,	315,	0,	"(") 		# function
  addCase(312,	rtne,	0,	0,	")") 		# end expr
  addCase(312,	emit,	300,	0,	"+") 
  addCase(312,	emit,	300,	0,	"-") 
  addCase(312,	emit,	300,	0,	"/") 
  addCase(312,	emit,	300,	0,	"?") 
  addCase(312,	emit,	300,	0,	"*") 
  addCase(312,	emit,	300,	0,	"=") 		# assignment
  addCase(312,	emit,	300,	0,	".") 	
  addCase(312,	emit,	338,	0,	"order") 	
  addCase(312,	drop,	310,	UNSP,	"collate") 	
  addCase(312,	skip,	0,	0,	"<key>") 	
  addCase(312,	drop,	312,	UNSP,	"<value>") 	

  addCase(313,	skip,	0,	0,	"<any>") 	

  addCase(314,	isrt,	320,	0,	"from") 	

  addCase(315,	rtne,	0,	0,	")") 		# end function
  addCase(315,	call,	315,	316,	"(") 		
  addCase(315,	emit,	315,	0,	"-")
  addCase(315,	call,	300,	0,	"select") 
  addCase(315,	cald,	1110,	316,	"convert") 	
  addCase(315,	emit,	303,	UNSP,	"datepart") 
  addCase(315,	emit,	315,	0,	"distinct") 
  addCase(315,	repl,	316,	INFO,	"to") 	
  addCase(315,	repl,	316,	0,	"char") 	
  addCase(315,	repl,	316,	0,	"bit") 	
  addCase(315,	emit,	316,	0,	"<value>") 

  addCase(316,	emit,	315,	0,	",") 
  addCase(316,	emit,	315,	0,	".") 
  addCase(316,	call,	315,	0,	"(") 
  addCase(316,	rtne,	0,	0,	")") 		# end function
  addCase(316,	emit,	316,	0,	"sum")
  addCase(316,	prto,	320,	0,	"from")
  addCase(316,	call,	360,	0,	"case") 		
  addCase(316,	rtne,	0,	0,	"end") 		# from case ?
  addCase(316,	emit,	315,	0,	"as")
  addCase(316,	emit,	315,	0,	"/")
  addCase(316,	emit,	315,	0,	"?")
  addCase(316,	emit,	315,	0,	"*")
  addCase(316,	emit,	315,	0,	"-")
  addCase(316,	emit,	315,	0,	"+")
  addCase(316,	skip,	0,	0,	"<any>")

  addCase(320,	emit,	320,	0,	"from") 	# for into
  addCase(320,	call,	320,	0,	"(")
  addCase(320,	call,	300,	0,	"select")
  addCase(320,	emit,	355,	0,	"union")
  addCase(320,	drop,	320,	SERR,	".")
  addCase(320,	emit,	320,	0,	"all")
  addCase(320,	emit,	322,	0,	"sysobjects")
  addCase(320,	emit,	360,	0,	"sysindexes")
  addCase(320,	emit,	2400,	0,	"information_schema")
  addCase(320,	skip,	0,	0,	"<key>")
  addCase(320,	emit,	330,	0,	"<value>") 	# from name

  addCase(322,	emit,	323,	0,	"where")
  addCase(322,	jump,	330,	0,	"<any>")
  addCase(323,	repl,	324,	0,	"name")
  addCase(323,	emit,	328,	0,	"type")
  addCase(323,	emit,	328,	0,	"xtype")
  addCase(323,	jump,	340,	0,	"<any>")
  addCase(324,	emit,	325,	0,	"=")
  addCase(325,	isrt,	326,	0,	"object_id(")
  addCase(326,	emit,	327,	0,	"<value>")
  addCase(327,	isrt,	354,	0,	")")

  addCase(328,	emit,	329,	0,	"=")
  addCase(328,	emit,	323,	0,	"and")

  addCase(329,	emit,	328,	0,	"<value>")

  addCase(330,	emit,	320,	0,	".") 		# schema
  addCase(330,	emit,	320,	0,	",") 		# list
  addCase(330,	emit,	320,	0,	"as") 		
  addCase(330,	emit,	330,	0,	"inner") 	
  addCase(330,	emit,	330,	0,	"outer") 	
  addCase(330,	emit,	330,	0,	"join") 	
  addCase(330,	emit,	330,	0,	"left") 	
  addCase(330,	repl,	330,	INFO,	"to") 	
  addCase(330,	emit,	340,	0,	"where") 	
  addCase(330,	emit,	320,	0,	"union") 	
  addCase(330,	emit,	335,	0,	"on") 	
  addCase(330,	emit,	345,	0,	"group") 	
  addCase(330,	emit,	338,	0,	"order") 	
  addCase(330,	call,	330,	0,	"(") 	
  addCase(330,	rtne,	0,	0,	")") 	
  addCase(330,	skip,	0,	0,	"<key>") 	
  addCase(330,	emit,	330,	0,	"<value>") 	# alias

  addCase(335,	call,	335,	336,	"(") 		# on
  addCase(335,	repl,	336,	INFO,	"to") 	
  addCase(335,	emit,	336,	0,	"<value>") 

  addCase(336,	emit,	335,	0,	".") 
  addCase(336,	emit,	320,	0,	",") 
  addCase(336,	emit,	335,	0,	"=") 
  addCase(336,	emit,	335,	0,	"on") 
  addCase(336,	emit,	340,	0,	"where") 	
  addCase(336,	emit,	355,	0,	"union") 	
  addCase(336,	emit,	341,	0,	"all") 	
  addCase(336,	rtne,	0,	0,	")") 
  addCase(336,	emit,	336,	0,	"inner") 	
  addCase(336,	emit,	336,	0,	"outer") 	
  addCase(336,	emit,	330,	0,	"join") 	
  addCase(336,	emit,	336,	0,	"left") 	
  addCase(336,	emit,	336,	0,	"right") 	
  addCase(336,	emit,	335,	0,	"and") 	
  addCase(336,	call,	338,	337,	"order") 	
  addCase(336,	call,	345,	337,	"group") 	
  addCase(336,	skip,	0,	0,	"<any>") 	

  addCase(337,	call,	345,	0,	"group") 	
  addCase(337,	skip,	0,	0,	"<any>") 	

  addCase(338,	emit,	339,	0,	"by") 		# order by
  addCase(338,	emit,	338,	0,	"desc") 
  addCase(338,	emit,	338,	0,	"asc") 
  addCase(338,	emit,	339,	0,	"as") 
  addCase(338,	emit,	339,	0,	",") 	
  addCase(338,	emit,	339,	0,	".") 	
  addCase(338,	emit,	339,	0,	"+") 	
  addCase(338,	drop,	338,	0,	"for")
  addCase(338,	drop,	338,	UNSP,	"xml")
  addCase(338,	drop,	338,	0,	"explicit")
  addCase(338,	call,	339,	0,	"(") 	
  addCase(338,	rtne,	0,	0,	")") 	
  addCase(338,	skip,	0,	0,	"<key>") 	
  addCase(338,	drop,	338,	UNSP,	"<value>") 	

  addCase(339,	emit,	338,	0,	"<value>") 	

  addCase(340,	call,	340,	350,	"(") 	
  addCase(340,	call,	300,	350,	"select") 	
  addCase(340,	emit,	340,	0,	"not") 	
  addCase(340,	emit,	340,	0,	"or") 	
  addCase(340,	emit,	340,	0,	"and") 	
  addCase(340,	emit,	340,	0,	">") 	
  addCase(340,	emit,	340,	0,	"!") 	
  addCase(340,	emit,	340,	0,	"exists") 	
  addCase(340,	emit,	340,	0,	"having") 	
  addCase(340,	emit,	338,	0,	"order") 	
  addCase(340,	repl,	350,	INFO,	"to") 	
  addCase(340,	skip,	0,	0,	"<key>") 	
  addCase(340,	emit,	350,	0,	"<value>") 	# where column

  addCase(341,	call,	300,	337,	"select") 	

  addCase(345,	emit,	346,	0,	"by") 		# group by
  addCase(346,	call,	346,	348,	"(") 		
  addCase(346,	call,	360,	348,	"case") 		
  addCase(346,	skip,	0,	0,	"<key>") 		
  addCase(346,	repl,	348,	0,	"offset")
  addCase(346,	emit,	348,	0,	"<value>")

  addCase(348,	emit,	346,	0,	"/")
  addCase(348,	emit,	346,	0,	"?")
  addCase(348,	emit,	346,	0,	"*")
  addCase(348,	emit,	346,	0,	"+")
  addCase(348,	emit,	346,	0,	"-")
  addCase(348,	emit,	346,	0,	",")
  addCase(348,	emit,	346,	0,	".")
  addCase(348,	call,	346,	348,	"(") 		
  addCase(348,	drop,	348,	0,	"with")
  addCase(348,	drop,	348,	UNSP,	"rollup")
  addCase(348,	emit,	340,	0,	"having")
  addCase(348,	emit,	338,	0,	"order") 	
  addCase(348,	rtne,	0,	0,	")")
  addCase(348,	skip,	0,	0,	"<any>") 	

  addCase(349,	dlim,	0,	1,	"<any>") 	

  addCase(350,	emit,	352,	0,	"=") 		# operator
  addCase(350,	emit,	352,	0,	"/") 		# operator
  addCase(350,	emit,	352,	0,	"?") 		# operator
  addCase(350,	emit,	352,	0,	"*") 		# operator
  addCase(350,	emit,	352,	0,	"-") 		# operator
  addCase(350,	emit,	352,	0,	"+") 		# operator
  addCase(350,	emit,	352,	0,	"<") 		# operator
  addCase(350,	emit,	352,	0,	">") 		# operator
  addCase(350,	emit,	340,	0,	".") 		# operator
  addCase(350,	emit,	340,	0,	"like") 	
  addCase(350,	emit,	350,	0,	"not") 	
  addCase(350,	emit,	350,	0,	"exists") 	
  addCase(350,	emit,	350,	0,	"!") 	
  addCase(350,	emit,	352,	0,	"in") 	
  addCase(350,	emit,	352,	0,	"is") 	
  addCase(350,	emit,	352,	0,	"isnull") 	
  addCase(350,	emit,	340,	0,	"between") 	
  addCase(350,	emit,	340,	0,	"or") 	
  addCase(350,	emit,	340,	0,	"and") 	
  addCase(350,	emit,	355,	0,	"union") 	
  addCase(350,	call,	338,	0,	"order") 	
  addCase(350,	emit,	345,	0,	"group") 	
  addCase(350,	call,	356,	0,	"(") 		# function
  addCase(350,	rtne,	0,	0,	")") 		
  addCase(350,	skip,	0,	0,	"<any>") 		

  addCase(352,	call,	356,	354,	"(") 		# function
  addCase(352,	emit,	352,	0,	">")
  addCase(352,	emit,	352,	0,	"=")
  addCase(352,	emit,	352,	0,	"-")
  addCase(352,	repl,	354,	0,	"0x0")
  addCase(352,	emit,	352,	0,	"not")
  addCase(352,	emit,	354,	0,	"null")
  addCase(352,	emit,	354,	0,	"<value>") 	# column

  addCase(354,	emit,	340,	0,	"and") 		# joiner
  addCase(354,	emit,	355,	0,	"union") 	
  addCase(354,	emit,	352,	0,	".") 	
  addCase(354,	emit,	340,	0,	"or") 	
  addCase(354,	call,	338,	0,	"order") 	
  addCase(354,	call,	345,	0,	"group") 	
  addCase(354,	emit,	354,	0,	"with") 	
  addCase(354,	emit,	354,	0,	"rollup") 	
  addCase(354,	call,	356,	0,	"(") 	
  addCase(354,	rtne,	0,	0,	")") 	
  addCase(354,	emit,	350,	0,	"not") 		
  addCase(354,	emit,	352,	0,	"+") 		
  addCase(354,	skip,	0,	0,	"<any>") 	

  addCase(355,	emit,	355,	0,	"all") 		# union
  addCase(355,	call,	355,	0,	"(") 	
  addCase(355,	rtne,	0,	0,	")") 	
  addCase(355,	call,	300,	354,	"select") 	
  addCase(355,	skip,	0,	0,	"<any>") 	

  addCase(356,	call,	300,	357,	"select") 	
  addCase(356,	emit,	356,	0,	"-") 	
  addCase(356,	emit,	357,	0,	"<value>") 	# plist value

  addCase(357,	emit,	356,	0,	",") 	
  addCase(357,	emit,	356,	0,	"*") 	
  addCase(357,	emit,	356,	0,	"-") 	
  addCase(357,	rtne,	0,	0,	")") 	
  addCase(357,	call,	356,	357,	"(") 	
  addCase(357,	skip,	0,	0,	"<any>") 	

  addCase(360,	call,	361,	362,	"(") 		# case
  addCase(360,	emit,	361,	0,	"when") 	 
  addCase(360,	emit,	361,	0,	"<value>") 	

  addCase(361,	emit,	360,	0,	".") 	 
  addCase(361,	emit,	361,	0,	"when") 	 
  addCase(361,	call,	300,	0,	"select") 	
  addCase(361,	cald,	1110,	362,	"convert") 	
  addCase(361,	call,	363,	362,	"(") 	
  addCase(361,	emit,	361,	0,	"-") 	
  addCase(361,	emit,	361,	0,	"=") 	
  addCase(361,	emit,	361,	0,	">") 	
  addCase(361,	call,	361,	362,	"case") 	
  addCase(361,	emit,	362,	0,	"<value>") 	

  addCase(362,	emit,	361,	0,	".") 	
  addCase(362,	emit,	361,	0,	",") 	
  addCase(362,	emit,	361,	0,	"-") 	
  addCase(362,	emit,	361,	0,	"+") 	
  addCase(362,	emit,	361,	0,	"/") 	
  addCase(362,	emit,	361,	0,	"?") 	
  addCase(362,	emit,	361,	0,	"*") 	
  addCase(362,	emit,	361,	0,	"=") 	
  addCase(362,	emit,	361,	0,	"<") 	
  addCase(362,	emit,	361,	0,	">") 	
  addCase(362,	emit,	362,	0,	"is") 	
  addCase(362,	emit,	361,	0,	"or") 	
  addCase(362,	emit,	362,	0,	"not") 	
  addCase(362,	emit,	362,	0,	"null") 	
  addCase(362,	emit,	361,	0,	"then") 	
  addCase(362,	emit,	361,	0,	"else") 	
  addCase(362,	emit,	361,	0,	"when") 	
  addCase(362,	call,	361,	0,	"case") 	
  addCase(362,	call,	363,	0,	"(") 	
  addCase(362,	rtne,	0,	0,	")") 	
  addCase(362,	rtne,	0,	0,	"end") 	
  addCase(362,	emit,	362,	0,	"in") 	
  addCase(362,	skip,	0,	0,	"<any>") 	

  addCase(363,	call,	300,	362,	"select") 	
  addCase(363,	cald,	1110,	362,	"convert") 	
  addCase(363,	call,	363,	364,	"(") 	
  addCase(363,	emit,	364,	0,	"<value>") 	# plist value

  addCase(364,	emit,	363,	0,	".") 	
  addCase(364,	emit,	363,	0,	",") 	
  addCase(364,	emit,	363,	0,	"/") 	
  addCase(364,	emit,	363,	0,	"?") 	
  addCase(364,	emit,	363,	0,	"*") 	
  addCase(364,	emit,	363,	0,	"+") 	
  addCase(364,	emit,	363,	0,	"=") 	
  addCase(364,	emit,	363,	0,	"&") 	
  addCase(364,	emit,	363,	0,	"as")
  addCase(364,	emit,	363,	0,	"-") 	
  addCase(364,	rtne,	0,	0,	")") 	
  addCase(364,	call,	363,	0,	"(") 	
  addCase(364,	rtne,	0,	0,	"end") 	

  addCase(370,	emit,	373,	0,	"where")	# sysindexes processing
  addCase(370,	jump,	330,	0,	"<any>")
  addCase(373,	emit,	374,	0,	"name")
  addCase(373,	jump,	340,	0,	"<any>")
  addCase(374,	emit,	375,	0,	"=")
  addCase(376,	lowr,	354,	0,	"<value>")

# Raiseerror 

  addCase(400,	cald,	420,	0,	"(") 	
  addCase(400,	drop,	400,	0,	"with") 	
  addCase(400,	rtnd,	0,	0,	"log") 	
  addCase(400,	skip,	0,	0,	"<any>") 	

  addCase(420,	emit,	430,	0,	"<value>") 	

  addCase(430,	emit,	430,	0,	",") 	
  addCase(430,	rtnd,	0,	0,	")") 	
  addCase(430,	emit,	430,	0,	"<value>") 	

# Drop

  addCase(500,	emit,	501,	0,	"table") 	
  addCase(500,	emit,	503,	0,	"view") 	
  addCase(500,	emit,	501,	0,	"constraint") 	
  addCase(500,	emit,	505,	0,	"function") 	
  addCase(500,	repl,	505,	0,	"procedure") 	
  addCase(500,	repl,	510,	0,	"trigger") 	

  addCase(501,	dtab,	502,	0,	"<value>") 	# Table name
  addCase(502,	emit,	502,	0,	"cascade") 	
  addCase(502,	emit,	501,	0,	".") 	
  addCase(502,	skip,	0,	0,	"<any>") 	

  addCase(503,	emit,	504,	0,	"if") 	
  addCase(503,	emit,	502,	0,	"<value>") 	# View name

  addCase(504,	emit,	503,	0,	"exists") 	

  addCase(505,	elog,	506,	0,	"<value>") 	# Function/Proc
  addCase(506,	isig,	502,	1,	"<any>") 	

  addCase(510,	elog,	512,	0,	"<value>") 	# Trigger
  addCase(512,	isrt,	514,	0,	"on") 	
  addCase(514,	tprt,	515,	0,	"<any>") 	
  addCase(515,	isrt,	516,	0,	"; drop rule if exists") 	
  addCase(516,	eprt,	517,	0,	"<any>") 	
  addCase(517,	isrt,	518,	0,	"on") 	
  addCase(518,	tprt,	519,	0,	"<any>") 	

  addCase(519,	isrt,	520,	0,	"; drop function") 	
  addCase(520,	eprt,	506,	0,	"<any>") 	
  
# Truncate

  addCase(550,	emit,	551,	0,	"table") 	
  addCase(551,	rtne,	0,	0,	"<value>") 	

# Print

  addCase(600,	emit,	610,	0,	"cast") 	
  addCase(600,	emit,	610,	0,	"<value>") 	
  addCase(610,	call,	620,	0,	"(") 	
  addCase(610,	rtne,	0,	0,	")") 	
  addCase(610,	emit,	620,	0,	"as") 	
  addCase(610,	emit,	620,	0,	"+") 	
  addCase(610,	skip,	0,	0,	"<any>") 	
  
  addCase(620,	emit,	610,	0,	"<value>") 	

# If statement

  addCase(750,	calp,	700,	752,	"<any>") 	
  addCase(752,	isrt,	753,	0,	"then") 	
  addCase(753,	trap,	0,	755,	"<any>") 	

  addCase(755,	trap,	0,	755,	"else") 	
  addCase(755,	isrt,	760,	0,	"end if") 	
  addCase(760,	skip,	0,	0,	"<any>") 	

# If expression

  addCase(700,	call,	700,	720,	"(") 		
  addCase(700,	emit,	700,	0,	"not") 		
  addCase(700,	emit,	700,	0,	"or") 		
  addCase(700,	emit,	700,	0,	">") 		
  addCase(700,	emit,	700,	0,	"=") 		
  addCase(700,	emit,	700,	0,	"exists") 	
  addCase(700,	call,	300,	0,	"select") 	
  addCase(700,	skip,	0,	0,	"<key>") 	
  addCase(700,	emit,	720,	0,	"<value>") 	# value

  addCase(720,	emit,	700,	0,	".") 		
  addCase(720,	emit,	722,	0,	"<") 		# operator
  addCase(720,	emit,	722,	0,	">") 		
  addCase(720,	emit,	722,	0,	"=") 		
  addCase(720,	emit,	722,	0,	"-") 		
  addCase(720,	emit,	722,	0,	"+") 		
  addCase(720,	emit,	733,	0,	"%") 		
  addCase(720,	emit,	730,	0,	"in") 	
  addCase(720,	emit,	720,	0,	"!") 	
  addCase(720,	emit,	720,	0,	"is") 	
  addCase(720,	emit,	720,	0,	"not") 	
  addCase(720,	emit,	700,	0,	"or") 	
  addCase(720,	emit,	700,	0,	"and") 	
  addCase(720,	emit,	740,	0,	"null") 	
  addCase(720,	call,	722,	0,	"(") 		# function
  addCase(720,	rtne,	0,	0,	")") 	
  addCase(720,	skip,	0,	0,	"<any>") 	

  addCase(722,	emit,	722,	0,	">") 		# <>
  addCase(722,	emit,	722,	0,	"-") 	
  addCase(722,	emit,	722,	0,	"=") 		# <=
  addCase(722,	emit,	740,	0,	"<value>") 	# value

  addCase(730,	emit,	732,	0,	"(") 		# list
  addCase(730,	rtne,	0,	0,	")") 	

  addCase(732,	emit,	734,	0,	"<value>") 	
  addCase(733,	emit,	720,	0,	"<value>") 	

  addCase(734,	emit,	732,	0,	",") 	
  addCase(734,	rtne,	0,	0,	")") 	
  addCase(734,	emit,	734,	0,	"<value>") 	# <value> <value>

  addCase(740,	emit,	700,	0,	"and") 		# joiner
  addCase(740,	emit,	700,	0,	"or") 		# joiner
  addCase(740,	emit,	722,	0,	",") 		
  addCase(740,	emit,	722,	0,	"+") 		
  addCase(740,	emit,	722,	0,	"-") 		
  addCase(740,	rtne,	0,	0,	")") 	
  addCase(740,	call,	722,	0,	"(") 		# function
  addCase(740,	skip,	0,	0,	"<any>") 	

# Declare 

  addCase(800,	save,	805,	0,	"<value>") 		# varname

  addCase(805,	drop,	815,	0,	"cursor") 	
  addCase(805,	drop,	806,	0,	"table") 	
  addCase(805,	jump,	809,	0,	"<any>") 	

  addCase(806,	isrb,	807,	0,	"bogus int; begin create temp table") 	
  addCase(807,	load,	115,	0,	"<any>") 	

  addCase(809,	load,	810,	0,	"<any>") 	

  addCase(810,	rtne,	0,	0,	"sysname") 	
  addCase(810,	emit,	820,	0,	"cursor") 	
  addCase(810,	repl,	830,	0,	"nvarchar") 	
  addCase(810,	emit,	830,	0,	"varchar") 	
  addCase(810,	emit,	810,	0,	"int") 	
  addCase(810,	emit,	810,	0,	"bigint") 	
  addCase(810,	emit,	810,	0,	"smallint") 	
  addCase(810,	emit,	810,	0,	"real") 	
  addCase(810,	emit,	810,	0,	"float") 	
  addCase(810,	emit,	810,	0,	"char") 	
  addCase(810,	emit,	810,	0,	"nchar") 	
  addCase(810,	repl,	810,	0,	"tinyint") 	
  addCase(810,	repl,	810,	0,	"money") 	
  addCase(810,	repl,	810,	0,	"bit") 	
  addCase(810,	call,	800,	0,	"(") 	
  addCase(810,	rtne,	0,	0,	")") 	
  addCase(810,	drop,	810,	0,	"as") 	
  addCase(810,	repl,	810,	0,	"datetime") 	
  addCase(810,	repl,	810,	0,	"uniqueidentifier") 	
  addCase(810,	repl,	800,	0,	",") 	
  addCase(810,	skip,	0,	0,	"<any>") 	

  addCase(815,	load,	817,	0,	"<any>") 	
  addCase(817,	isrt,	820,	0,	"cursor") 	

  addCase(820,	drop,	820,	0,	"local") 		# no hold 
  addCase(820,	drop,	820,	0,	"forward_only") 	# no scoll	
  addCase(820,	drop,	820,	0,	"fast_forward") 	
  addCase(820,	drop,	820,	0,	"static") 		
  addCase(820,	drop,	820,	0,	"read_only") 	
  addCase(820,	emit,	821,	0,	"for") 	

  addCase(821,	drop,	821,	0,	"begin") 	
  addCase(821,	emit,	300,	0,	"select") 	

  addCase(830,	call,	832,	810,	"(") 		# varchar
  addCase(830,	rtne,	0,	0,	")") 	
  addCase(832,	emit,	830,	0,	"<value>") 	

# Open

  addCase(900,	rtne,	0,	0,	"<value>") 	

# Deallocate

  addCase(910,	rtnd,	920,	0,	"<value>") 	

# Fetch

  addCase(1000,	emit,	1000,	0,	"next") 	
  addCase(1000,	emit,	1010,	0,	"from") 	
  addCase(1000,	emit,	1010,	0,	"into") 	
  addCase(1000,	emit,	1010,	0,	",") 	
  addCase(1000,	skip,	0,	0,	"<any>") 	

  addCase(1010,	emit,	1000,	0,	"<value>") 	

# Set

  addCase(1100,	drop,	1130,	0,	"transaction") 	
  addCase(1100,	drop,	1141,	0,	"ansi_nulls") 	
  addCase(1100,	drop,	1141,	INFO,	"ansi_padding") 	
  addCase(1100,	drop,	1140,	INFO,	"quoted_identifier") 	
  addCase(1100,	drop,	1143,	0,	"dateformat") 	
  addCase(1100,	drop,	1140,	0,	"nocount") 	
  addCase(1100,	drop,	1142,	0,	"identity_insert") 	
  addCase(1100,	drop,	1140,	INFO,	"xact_abort") 	

  addCase(1100,	drop,	1145,	UNSP,	"ansi_warnings") 	

  addCase(1100,	drop,	1140,	UNSP,	"rowcount") 		
  addCase(1100,	drop,	1140,	INFO,	"ansi_null_default") 	
  addCase(1100,	drop,	1140,	INFO,	"concat_null_yields_null") 	
  addCase(1100,	drop,	1140,	INFO,	"numeric_roundabort") 	
  addCase(1100,	drop,	1140,	INFO,	"recursive_triggers") 	
  addCase(1100,	drop,	1140,	INFO,	"auto_close") 	
  addCase(1100,	drop,	1140,	INFO,	"auto_shrink") 	
  addCase(1100,	drop,	1140,	INFO,	"auto_create_statistics") 	
  addCase(1100,	drop,	1140,	INFO,	"auto_update_statistics") 	
  addCase(1100,	drop,	1140,	INFO,	"cursor_close_on_commit") 	
  addCase(1100,	drop,	1140,	INFO,	"cursor_default") 	
  addCase(1100,	drop,	1140,	INFO,	"arithignore") 	
  addCase(1100,	drop,	1140,	INFO,	"arithabort") 	
  addCase(1100,	drop,	1140,	INFO,	"recovery") 	
  addCase(1100,	repr,	0,	INFO,	"read_write") 	
  addCase(1100,	repr,	0,	INFO,	"multi_user") 	

  addCase(1100,	skip,	0,	0,	"<key>") 	
  addCase(1100,	emit,	1102,	0,	"<value>") 	

  addCase(1102,	emit,	1100,	0,	".") 		
  addCase(1102,	repl,	1104,	0,	"=") 	
  addCase(1102,	rtne,	1104,	0,	"off") 	
  addCase(1102,	rtne,	1104,	0,	"on") 	

  addCase(1104,	call,	1105,	1106,	"(") 	
  addCase(1104,	call,	360,	0,	"case") 	
  addCase(1104,	cald,	1110,	0,	"convert") 	
  addCase(1104,	skip,	0,	0,	"<key>") 	
  addCase(1104,	emit,	1106,	0,	"<value>") 	

  addCase(1105,	call,	300,	1106,	"select") 	
  addCase(1105,	cald,	1110,	1106,	"convert") 	
  addCase(1105,	call,	1105,	1106,	"(") 	
  addCase(1105,	skip,	0,	0,	"<key>") 	
  addCase(1105,	emit,	1106,	0,	"<value>") 	

  addCase(1106,	call,	1104,	0,	"(") 	
  addCase(1106,	rtne,	0,	0,	")") 	
  addCase(1106,	emit,	1104,	0,	",") 	
  addCase(1106,	emit,	1104,	0,	".") 	
  addCase(1106,	emit,	1104,	0,	"+") 	
  addCase(1106,	emit,	1104,	0,	"-") 	
  addCase(1106,	emit,	1104,	0,	"/") 	
  addCase(1106,	emit,	1104,	0,	"?") 	
  addCase(1106,	emit,	1104,	0,	"*") 	
  addCase(1106,	emit,	1104,	0,	"as")
  addCase(1106,	emit,	320,	0,	"from")
  addCase(1106,	emit,	1104,	0,	"=")
  addCase(1106,	skip,	0,	0,	"<key>") 	
  addCase(1106,	emit,	1106,	0,	"<value>") 	

  addCase(1110,	isrt,	1111,	0,	"cast") 	# convert
  addCase(1111,	emit,	1112,	0,	"(") 	
  addCase(1112,	reps,	1113,	0,	"nvarchar") 	
  addCase(1112,	reps,	1113,	0,	"uniqueidentifier") 	
  addCase(1112,	reps,	1113,	0,	"bit") 	
  addCase(1112,	save,	1113,	0,	"<value>") 	
  addCase(1113,	save,	1114,	0,	"(") 	
  addCase(1113,	drop,	1117,	0,	",") 	
  addCase(1114,	save,	1115,	0,	"<value>") 	
  addCase(1115,	save,	1116,	0,	")") 	
  addCase(1116,	drop,	1117,	0,	",") 	

  addCase(1117,	drop,	1118,	0,	")") 		# closing bracket
  addCase(1117,	drop,	1123,	0,	",") 		# extra parameter	
  addCase(1117,	call,	1122,	0,	"(") 	
  addCase(1117,	emit,	1117,	0,	"<value>") 	
  addCase(1118,	isrt,	1119,	0,	"as") 	
  addCase(1119,	loda,	1120,	0,	"<any>") 	
  addCase(1120,	isrt,	1121,	0,	")") 	
  addCase(1121,	skip,	0,	0,	"<any>") 	

  addCase(1122,	rtne,	0,	0,	")") 	
  addCase(1122,	call,	1122,	0,	"(") 	
  addCase(1122,	emit,	1122,	0,	"<value>") 	

  addCase(1123,	drop,	1117,	UNSP,	"<value>") 	# date format number

# Transaction

  addCase(1130,	isrt,	1131,	0,	"set transaction") 	
  addCase(1131,	skip,	0,	0,	"<key>") 	
  addCase(1131,	emit,	1131,	0,	"<value>") 	

# System variables

  addCase(1140,	isrt,	1150,	0,	"set escape_string_warning") 	# catch all
  addCase(1141,	isrt,	1150,	0,	"set transform_null_equals") 	
  addCase(1142,	isrt,	1154,	0,	"set escape_string_warning") 	
  addCase(1143,	isrt,	1150,	0,	"set datestyle") 	
  addCase(1144,	isrt,	1145,	0,	"set escape_string_warning") 	

  addCase(1145,	ansi,	1150,	1,	"on") 	
  addCase(1145,	ansi,	1150,	0,	"off") 	

  addCase(1150,	repr,	0,	0,	"on") 	
  addCase(1150,	repr,	0,	0,	"off") 	
  addCase(1150,	repr,	0,	0,	"mdy") 	
  addCase(1150,	repr,	0,	0,	"global") 	
  addCase(1150,	repr,	0,	0,	"simple") 	
  addCase(1150,	repr,	0,	0,	"1") 	
  addCase(1150,	repr,	0,	0,	"0") 	
  addCase(1150,	skip,	0,	0,	"<key>") 	
  addCase(1150,	drop,	1151,	0,	"<value>") 	
  addCase(1151,	isrt,	1152,	0,	"=0") 	
  addCase(1152,	skip,	0,	0,	"<any>") 	

  addCase(1154,	drop,	1150,	0,	"<value>") 	

# While

  addCase(1250,	calp,	1200,	1252,	"<any>") 	
  addCase(1252,	isrt,	1253,	0,	"loop") 	
  addCase(1253,	trap,	0,	1255,	"<any>") 	

  addCase(1255,	isrt,	1260,	0,	"end loop") 	
  addCase(1260,	skip,	0,	0,	"<any>") 	

  addCase(1200,	call,	1200,	0,	"(") 	
  addCase(1200,	skip,	0,	0,	"<key>") 	
  addCase(1200,	drop,	1205,	0,	"__fetch_status") 	
  addCase(1200,	emit,	1210,	0,	"<value>") 	

  addCase(1205,	drop,	1206,	0,	"=") 	
  addCase(1206,	repr,	0,	0,	"0") 	

  addCase(1210,	emit,	1220,	0,	"=") 	
  addCase(1210,	emit,	1200,	0,	",") 	
  addCase(1210,	call,	1200,	0,	"(") 	
  addCase(1210,	rtne,	0,	0,	")") 	
  addCase(1210,	emit,	1220,	0,	"<") 	
  addCase(1210,	emit,	1220,	0,	">") 	
  addCase(1210,	emit,	1210,	0,	"is") 	
  addCase(1210,	emit,	1210,	0,	"not") 	
  addCase(1210,	emit,	1210,	0,	"null") 	
  addCase(1210,	skip,	0,	0,	"<any>") 	

  addCase(1220,	emit,	1220,	0,	"=") 	
  addCase(1220,	emit,	1220,	0,	">") 	
  addCase(1220,	skip,	0,	0,	"<key>") 	
  addCase(1220,	emit,	1210,	0,	"<value>") 	

# Execute

  addCase(1300,	isrt,	1301,	0,	"perform") 	

  addCase(1301,	emit,	1302,	0,	"(") 	
  addCase(1301,	emit,	1305,	0,	"sp_executesql") 	
  addCase(1301,	emit,	1310,	0,	"<value>") 	# function exec
 
  addCase(1302,	emit,	1303,	0,	"<value>") 	# string exec

  addCase(1303,	emit,	1302,	0,	"+") 	
  addCase(1303,	rtne,	0,	0,	")") 	
 
  addCase(1305,	isrt,	1306,	0,	"(") 	

  addCase(1306,	drop,	1306,	0,	"_statement") 	
  addCase(1306,	emit,	1307,	0,	"_sql") 	
  addCase(1306,	drop,	1306,	0,	"=") 	
  addCase(1306,	dump,	1307,	0,	"<value>") 	# string exec

  addCase(1307,	isrt,	1308,	0,	")") 	
  addCase(1308,	skip,	0,	0,	"<any>") 	

  addCase(1310,	emit,	1301,	0,	"=") 	
  addCase(1310,	emit,	1301,	0,	".") 	
  addCase(1310,	repl,	1311,	0,	"go")
  addCase(1310,	emit,	1312,	0,	"(") 	
  addCase(1310,	isrt,	1312,	0,	"(") 	

  addCase(1311,	skip,	0,	0,	"<any>") 	

  addCase(1312,	emit,	1314,	0,	"<value>") 	

  addCase(1314,	drop,	1314,	UNSP,	"output")
  addCase(1314,	emit,	1312,	0,	",") 	
  addCase(1314,	emit,	1312,	0,	"=") 	
  addCase(1314,	rtne,	0,	0,	")") 	
  addCase(1314,	isrt,	1315,	0,	")") 	

  addCase(1315,	skip,	0,	0,	"<any>") 	

  # Insert  

  addCase(1400,	emit,	1401,	0,	"into") 		
  addCase(1400,	isrt,	1401,	0,	"into") 		
  addCase(1401,	drop,	1401,	SERR,	".") 		
  addCase(1401,	emit,	1402,	0,	"<value>") 	# table name 

  addCase(1402,	emit,	1401,	0,	".") 		
  addCase(1402,	emit,	1404,	0,	"values") 		
  addCase(1402,	call,	300,	1405,	"select") 		
  addCase(1402,	call,	1403,	1404,	"(") 		
  addCase(1402,	repl,	1408,	0,	"exec") 		
  addCase(1402,	skip,	0,	0,	"<any>") 	

  addCase(1403,	call,	300,	1405,	"select") 	# column list
  addCase(1403,	drop,	1403,	SERR,	".") 	
  addCase(1403,	repl,	1404,	0,	"user") 	
  addCase(1403,	repl,	1404,	0,	"offset") 	
  addCase(1403,	emit,	1404,	0,	"<value>") 	

  addCase(1404,	emit,	1403,	0,	",") 		
  addCase(1404,	rtne,	0,	0,	")") 		
  addCase(1404,	emit,	1404,	0,	"-") 		# value list
  addCase(1404,	call,	300,	1405,	"select") 
  addCase(1404,	emit,	1404,	0,	"values") 		
  addCase(1404,	drop,	1404,	SERR,	".") 		
  addCase(1404,	call,	1404,	1405,	"(") 
  addCase(1404,	repl,	1408,	0,	"exec") 		
  addCase(1404,	call,	361,	1405,	"case") 	
  addCase(1404,	skip,	0,	0,	"<key>") 
  addCase(1404,	escp,	1405,	0,	"<value>") 

  addCase(1405,	emit,	1404,	0,	",") 		# operator
  addCase(1405,	emit,	1404,	0,	"*") 
  addCase(1405,	emit,	1404,	0,	"/") 
  addCase(1405,	emit,	1404,	0,	"?") 
  addCase(1405,	emit,	1404,	0,	"+") 
  addCase(1405,	emit,	1404,	0,	"-") 
  addCase(1405,	emit,	1404,	0,	".") 
  addCase(1405,	emit,	1404,	0,	"as") 
  addCase(1405,	call,	1404,	0,	"(") 		# function
  addCase(1405,	rtne,	0,	0,	")") 	
  addCase(1405,	skip,	0,	0,	"<any>") 	
  
  addCase(1408,	emit,	1409,	0,	"<value>") 	# exec function
  addCase(1409,	emit,	1408,	0,	".") 	
  addCase(1409,	emit,	1410,	0,	"(") 	
  addCase(1409,	isrt,	1410,	0,	"(") 	

  addCase(1410,	emit,	1412,	0,	"<value>") 	
  addCase(1412,	emit,	1410,	0,	",") 	
  addCase(1412,	emit,	1414,	0,	")") 	
  addCase(1412,	isrt,	1414,	0,	")") 	

  addCase(1414,	skip,	0,	0,	"<any>")

# Return 

  addCase(1500,	drop,	1505,	0,	"select") 	
  addCase(1500,	cald,	1110,	1510,	"convert") 	
  addCase(1500,	call,	1500,	1510,	"(") 	
  addCase(1500,	emit,	1500,	0,	"-") 	
  addCase(1500,	tsto,	0,	1,	"<key>") 	
  addCase(1500,	tsto,	1510,	0,	"<value>") 	

  addCase(1505,	isrt,	1506,	0,	"query select") 	
  addCase(1506,	calp,	300,	1510,	"<any>") 	

  addCase(1510,	rtne,	0,	0,	")") 	
  addCase(1510,	call,	1500,	0,	"(") 	
  addCase(1510,	emit,	1500,	0,	",") 	
  addCase(1510,	emit,	1500,	0,	".") 	
  addCase(1510,	emit,	1500,	0,	"/") 	
  addCase(1510,	emit,	1500,	0,	"?") 	
  addCase(1510,	emit,	1500,	0,	"+") 	
  addCase(1510,	emit,	1500,	0,	"-") 	
  addCase(1510,	emit,	1500,	0,	"*") 	
  addCase(1510,	emit,	1500,	0,	"%") 	
  addCase(1510,	emit,	1500,	0,	"=") 	
  addCase(1510,	emit,	1500,	0,	"or") 	
  addCase(1510,	emit,	1500,	0,	"and") 	
  addCase(1510,	emit,	1510,	0,	"is") 	
  addCase(1510,	emit,	1510,	0,	"not") 	
  addCase(1510,	emit,	1510,	0,	"null") 	
  addCase(1510,	skip,	0,	0,	"<any>") 	

# Update 

  addCase(1600,	emit,	1610,	0,	"<value>") 	# table

  addCase(1610,	emit,	1600,	0,	".") 	
  addCase(1610,	call,	1615,	0,	"(") 	
  addCase(1610,	emit,	1610,	0,	"set") 	
  addCase(1610,	repl,	1612,	0,	"user") 	
  addCase(1610,	repl,	1612,	0,	"offset") 	
  addCase(1610,	skip,	0,	0,	"<key>") 	
  addCase(1610,	emit,	1612,	0,	"<value>") 	

  addCase(1612,	emit,	1610,	0,	".") 	
  addCase(1612,	emit,	1615,	0,	"=") 	
  addCase(1612,	emit,	1640,	0,	"where") 	
  addCase(1612,	skip,	0,	0,	"<key>") 	
  addCase(1612,	emit,	1612,	0,	"<value>") 	# alias

  addCase(1615,	call,	1635,	0,	"(") 	
  addCase(1615,	call,	300,	1635,	"select") 	
  addCase(1615,	call,	361,	1635,	"case") 	
  addCase(1615,	skip,	0,	0,	"<key>") 	
  addCase(1615,	emit,	1635,	0,	"<value>") 	

  addCase(1620,	call,	1640,	0,	"(") 	
  addCase(1620,	emit,	1621,	0,	"<value>") 	
#  addCase(1621,	drop,	1622,	0,	"inner") 	# join
  addCase(1621,	emit,	1621,	0,	"inner") 	
  addCase(1621,	emit,	1620,	0,	"join") 	
  addCase(1621,	emit,	1640,	0,	"on") 	

  addCase(1621,	emit,	1640,	0,	"where") 	
  addCase(1621,	emit,	1620,	0,	",") 	
  addCase(1621,	skip,	0,	0,	"<key>") 	
  addCase(1621,	emit,	1621,	0,	"<value>") 	# alias

#  addCase(1622,	repl,	1623,	0,	"join") 	
#  addCase(1623,	call,	1626,	0,	"(") 	
#  addCase(1623,	emit,	1625,	0,	"<value>") 	
#  addCase(1625,	repl,	1640,	0,	"on") 	
#  addCase(1625,	emit,	1625,	0,	"<value>") 	# alias

#  addCase(1626,	emit,	300,	0,	"select") 	
#  addCase(1626,	rtne,	0,	0,	")") 	

  addCase(1635,	emit,	1610,	0,	",") 	
  addCase(1635,	emit,	1615,	0,	".") 	
  addCase(1635,	emit,	1615,	0,	"+") 	
  addCase(1635,	emit,	1615,	0,	"-") 	
  addCase(1635,	emit,	1615,	0,	"/") 	
  addCase(1635,	emit,	1615,	0,	"?") 	
  addCase(1635,	emit,	1615,	0,	"*") 	
  addCase(1635,	emit,	1620,	0,	"from") 	
  addCase(1635,	emit,	1640,	0,	"where") 	
  addCase(1635,	call,	1637,	0,	"(") 	
  addCase(1635,	rtne,	0,	0,	")") 	
  addCase(1635,	emit,	1640,	0,	"on") 	
  addCase(1635,	skip,	0,	0,	"<key>") 	
  addCase(1635,	emit,	1635,	0,	"<value>") 	

  addCase(1637,	rtne,	0,	0,	")") 	
  addCase(1637,	call,	1637,	1638,	"(") 	
  addCase(1637,	call,	300,	1635,	"select") 	
  addCase(1637,	emit,	1637,	0,	"-") 	
  addCase(1637,	emit,	1638,	0,	"<value>") 	# function parms

  addCase(1638,	emit,	1637,	0,	",") 	
  addCase(1638,	emit,	1637,	0,	".") 	
  addCase(1638,	emit,	1637,	0,	"/") 	
  addCase(1638,	emit,	1637,	0,	"?") 	
  addCase(1638,	emit,	1637,	0,	"?") 	
  addCase(1638,	emit,	1637,	0,	"+") 	
  addCase(1638,	emit,	1637,	0,	"-") 	
  addCase(1638,	emit,	1637,	0,	"*") 	
  addCase(1638,	emit,	1637,	0,	"as")
  addCase(1638,	rtne,	0,	0,	")") 	
  addCase(1638,	call,	1637,	1638,	"(") 	
  addCase(1638,	skip,	0,	0,	"<any>") 	

  addCase(1640,	emit,	1640,	0,	"current") 	
  addCase(1640,	emit,	1640,	0,	"of") 	
  addCase(1640,	emit,	300,	0,	"select") 	
  addCase(1640,	emit,	1640,	0,	"not") 	
  addCase(1640,	call,	1640,	1650,	"(") 	
  addCase(1640,	call,	300,	0,	"(") 	
  addCase(1640,	call,	361,	0,	"case") 	
  addCase(1640,	skip,	0,	0,	"<key>") 	
  addCase(1640,	emit,	1645,	0,	"<value>") 	

  addCase(1645,	emit,	1640,	0,	".") 	
  addCase(1645,	emit,	1640,	0,	",") 	
  addCase(1645,	emit,	1640,	0,	"like") 	
  addCase(1645,	emit,	1640,	0,	"as")
  addCase(1645,	emit,	1640,	0,	"or")
  addCase(1645,	emit,	1648,	0,	"in") 	
  addCase(1645,	emit,	1645,	0,	"is") 	
  addCase(1645,	emit,	1645,	0,	"not") 	
  addCase(1645,	emit,	1640,	0,	"and") 	
  addCase(1645,	emit,	1645,	0,	"exists") 	
  addCase(1645,	emit,	1640,	0,	"between") 	
  addCase(1645,	emit,	1650,	0,	"null") 	
  addCase(1645,	emit,	1647,	0,	"=") 	
  addCase(1645,	emit,	1640,	0,	"+") 	
  addCase(1645,	call,	1640,	0,	"(") 	
  addCase(1645,	rtne,	0,	0,	")") 	
  addCase(1645,	emit,	1647,	0,	"<") 	
  addCase(1645,	emit,	1647,	0,	">") 	
  addCase(1645,	skip,	0,	0,	"<any>") 	

  addCase(1647,	call,	1647,	1650,	"(") 	
  addCase(1647,	rtne,	0,	0,	")") 	
  addCase(1647,	emit,	1640,	0,	"and") 	
  addCase(1647,	emit,	1647,	0,	"=") 	
  addCase(1647,	emit,	1647,	0,	"-") 	
  addCase(1647,	emit,	1647,	0,	">") 	
  addCase(1647,	skip,	0,	0,	"<key>") 	
  addCase(1647,	emit,	1650,	0,	"<value>") 	
  
  addCase(1648,	call,	1648,	1650,	"(") 		# in
  addCase(1648,	call,	300,	1650,	"select") 	
  addCase(1648,	skip,	0,	0,	"<key>") 	
  addCase(1648,	emit,	1649,	0,	"<value>") 	

  addCase(1649,	emit,	1648,	0,	",") 	
  addCase(1649,	rtne,	0,	0,	")") 	

  addCase(1650,	rtne,	0,	0,	")") 	
  addCase(1650,	emit,	1640,	0,	"and") 	
  addCase(1650,	emit,	1640,	0,	"where") 	
  addCase(1650,	emit,	1640,	0,	"or") 	
  addCase(1650,	emit,	1652,	0,	"inner") 	
  addCase(1650,	emit,	1647,	0,	".") 	
  addCase(1650,	skip,	0,	0,	"<any>") 	

  addCase(1652,	emit,	1652,	0,	"join") 	
  addCase(1652,	emit,	1640,	0,	"on") 	
  addCase(1652,	emit,	1652,	0,	"<value>") 	

# Use

  addCase(1700,	drop,	1702,	UNSP,	"<value>") 	
  addCase(1702,	isrt,	1704,	0,	"null") 	
  addCase(1704,	skip,	0,	0,	"<any>") 	

# Begin

  addCase(1750,	drop,	1755,	0,	"begin") 	
  addCase(1755,	repl,	0,	0,	"transaction") 	
  addCase(1755,	repl,	0,	0,	"tran") 	
  addCase(1755,	isrt,	1760,	0,	"begin ")
  addCase(1760,	stab,	0,	0,	"<any>") 	

  addCase(1775,	emit,	1776,	0,	"end") 		# This will be end 
  addCase(1776,	endb,	1777,	0,	"<any>") 	# End of begin blocks
  addCase(1777,	skip,	0,	0,	"<any>") 

# Alter

  addCase(1800,	repl,	1810,	0,	"table") 	
  addCase(1800,	repl,	2000,	0,	"function") 	
  addCase(1800,	repl,	2000,	0,	"procedure") 	
  addCase(1800,	repl,	2200,	0,	"trigger") 	
  addCase(1800,	repl,	2300,	0,	"database") 	

  addCase(1810,	emit,	1820,	0,	"<value>") 	# alter table
  addCase(1820,	emit,	1810,	0,	".") 	
  addCase(1820,	drop,	1825,	0,	"with") 	
  addCase(1820,	drop,	1825,	0,	"check") 	
  addCase(1820,	emit,	1820,	0,	"constraint") 	
  addCase(1820,	drop,	1880,	0,	"add") 	
  addCase(1820,	emit,	1820,	0,	"drop") 	
  addCase(1820,	emit,	1823,	0,	"alter") 	
  addCase(1820,	emit,	1822,	0,	"column") 	
  addCase(1820,	skip,	0,	0,	"<key>") 	
  addCase(1820,	emit,	1830,	0,	"<value>") 	
 
  addCase(1822,	rtne,	0,	0,	"<value>") 	# alter column	

  addCase(1823,	emit,	2100,	0,	"column") 	# alter column 

  addCase(1825,	drop,	1820,	0,	"check") 	
  addCase(1825,	drop,	1820,	UNSP,	"nocheck") 	
  addCase(1825,	repl,	1826,	0,	"constraint") 	

  addCase(1826,	skip,	0,	0,	"<key>") 	
  addCase(1826,	drop,	1827,	0,	"<value>") 	
  addCase(1827,	skip,	0,	0,	"<any>") 	
 
  addCase(1830,	emit,	1830,	0,	"with") 	
  addCase(1830,	emit,	1830,	0,	"check") 	
  addCase(1830,	emit,	1830,	0,	"add") 	
  addCase(1830,	emit,	1840,	0,	"constraint") 	
  addCase(1830,	emit,	1890,	0,	"column") 	
  addCase(1830,	skip,	0,	0,	"<any>") 	

  addCase(1840,	emit,	1841,	0,	"<value>") 		# constraint
  addCase(1841,	emit,	1841,	0,	"primary") 	
  addCase(1841,	emit,	1841,	0,	"foreign") 	
  addCase(1841,	drop,	1841,	0,	"clustered") 	
  addCase(1841,	drop,	1841,	0,	"nonclustered") 	
  addCase(1841,	emit,	1841,	0,	"unique") 	
  addCase(1841,	drop,	1841,	0,	"asc") 	
  addCase(1841,	emit,	1850,	0,	"default") 	
  addCase(1841,	emit,	1841,	0,	"key") 	
  addCase(1841,	call,	1841,	0,	"(") 	
  addCase(1841,	rtne,	0,	0,	")") 	
  addCase(1841,	emit,	1860,	0,	"references") 	
  addCase(1841,	skip,	0,	0,	"<key>") 	
  addCase(1841,	emit,	1841,	0,	"<value>") 	
  
  addCase(1850,	emit,	1855,	0,	"(") 		# default
  addCase(1850,	emit,	1850,	0,	")") 	
  addCase(1850,	emit,	1855,	0,	"for") 	
  addCase(1850,	emit,	1856,	0,	"1") 	
  addCase(1850,	emit,	1856,	0,	"0") 	
  addCase(1850,	emit,	1830,	0,	",") 	
  addCase(1850,	skip,	0,	0,	"<any>") 	

  addCase(1855,	emit,	1855,	0,	"(") 		# default
  addCase(1855,	emit,	1855,	0,	"-") 	
  addCase(1855,	emit,	1856,	0,	"<value>") 	

  addCase(1856,	emit,	1856,	0,	")") 	
  addCase(1856,	emit,	1856,	0,	"(") 	
  addCase(1856,	emit,	1895,	0,	",") 	
  addCase(1856,	emit,	1857,	0,	"for") 	
  addCase(1856,	skip,	0,	0,	"<any>") 	
  
  addCase(1857,	emit,	1850,	0,	"<value>") 	

  addCase(1860,	emit,	1870,	0,	"<value>") 	# references
  addCase(1870,	emit,	1875,	0,	",") 	
  addCase(1870,	emit,	1860,	0,	".") 	
  addCase(1870,	call,	1875,	0,	"(") 	
  addCase(1870,	emit,	1870,	0,	"pkey") 	
  addCase(1870,	rtne,	0,	0,	")") 	
  addCase(1870,	emit,	1870,	0,	"on") 	
  addCase(1870,	emit,	1870,	0,	"update") 	
  addCase(1870,	emit,	1870,	0,	"delete") 	
  addCase(1870,	emit,	1870,	0,	"cascade") 	
  addCase(1870,	skip,	0,	0,	"<any>") 	
  addCase(1875,	emit,	1870,	0,	"<value>") 	# references

  addCase(1880,	drop,	1881,	0,	"constraint") 	# add constraint
  addCase(1880,	isrt,	1890,	0,	"add") 	
  addCase(1881,	save,	1882,	0,	"<value>") 	

  addCase(1882,	drop,	1883,	0,	"default") 	
  addCase(1882,	jump,	1896,	0,	"<any>") 	

  addCase(1883,	sclr,	1884,	0,	"<any>") 	
  addCase(1884,	drop,	1885,	0,	"for") 	
  addCase(1884,	save,	1884,	0,	"<value>") 	

  addCase(1885,	isrt,	1886,	0,	"alter column") 	
  addCase(1886,	emit,	1887,	0,	"<value>") 	
  addCase(1887,	isrt,	1888,	0,	"set default") 	
  addCase(1888,	loda,	1889,	0,	"<any>") 	
  addCase(1889,	emit,	1880,	0,	",") 	
  addCase(1889,	skip,	0,	0,	"<any>") 	

  addCase(1890,	emit,	1830,	0,	",") 	
  addCase(1890,	repl,	1891,	0,	"user") 	
  addCase(1890,	repl,	1891,	0,	"offset") 	
  addCase(1890,	emit,	1891,	0,	"<value>") 	# add column
  addCase(1891,	repl,	1892,	0,	"uniqueidentifier") 
  addCase(1891,	repl,	1892,	0,	"nvarchar") 
  addCase(1891,	repl,	1892,	0,	"ntext") 
  addCase(1891,	repl,	1892,	0,	"datetime") 
  addCase(1891,	repl,	1892,	0,	"money") 
  addCase(1891,	repl,	1892,	0,	"bit") 
  addCase(1891,	emit,	1892,	0,	"<value>") 	# add type

  addCase(1892,	repl,	1892,	0,	"bit") 	
  addCase(1892,	emit,	1892,	0,	"not") 	
  addCase(1892,	emit,	1892,	0,	"null") 	
  addCase(1892,	emit,	1894,	0,	"default") 	
  addCase(1892,	repl,	1894,	0,	"identity") 	
  addCase(1892,	drop,	1893,	UNSP,	"collate") 	
  addCase(1892,	call,	1894,	0,	"(") 	
  addCase(1892,	emit,	1894,	0,	"unique") 	
  addCase(1892,	emit,	1840,	0,	"constraint") 	
  addCase(1892,	emit,	1880,	0,	",") 	
  addCase(1892,	skip,	0,	0,	"<any>") 	

  addCase(1893,	drop,	1892,	0,	"<value>") 	# collate value 
  
  addCase(1894,	call,	1894,	0,	"(") 	
  addCase(1894,	rtne,	0,	0,	")") 	
  addCase(1894,	emit,	1892,	0,	"not")
  addCase(1894,	skip,	0,	0,	"<key>") 	
  addCase(1894,	emit,	1894,	0,	"<value>") 	

  addCase(1895,	emit,	1890,	0,	"add") 	
  addCase(1895,	isrt,	1890,	MSNG,	"add") 	

  addCase(1896,	isrt,	1897,	0,	"add constraint") 	
  addCase(1897,	load,	1841,	0,	"<any>") 	

# Delete

  addCase(1900,	emit,	1910,	0,	"from") 	
  addCase(1900,	drop,	1900,	UNSP,	"<value>") 	

  addCase(1910,	emit,	1920,	0,	"<value>") 		# table
  addCase(1920,	emit,	1910,	0,	".") 		

  addCase(1920,	emit,	1920,	0,	"inner") 		
  addCase(1920,	emit,	1920,	0,	"left") 		
  addCase(1920,	emit,	1920,	0,	"outer") 		
  addCase(1920,	emit,	1910,	0,	"join") 		
  addCase(1920,	emit,	1910,	0,	"on") 		

  addCase(1920,	emit,	1930,	0,	"where") 		
  addCase(1920,	emit,	1910,	0,	"=") 		
  addCase(1920,	repl,	1921,	0,	"from") 		
  addCase(1920,	repl,	1920,	0,	",") 		
  addCase(1920,	skip,	0,	0,	"<key>") 	
  addCase(1920,	emit,	1920,	0,	"<value>") 		# alias

  addCase(1921,	emit,	1922,	0,	"<value>") 		
  addCase(1922,	emit,	1922,	UNSP,	"left") 		
  addCase(1922,	emit,	1923,	0,	"inner") 		
  addCase(1922,	emit,	1923,	UNSP,	"outer") 		
  addCase(1923,	emit,	1924,	0,	"join") 		
  addCase(1924,	emit,	1925,	0,	"<value>") 		# using table
  addCase(1925,	emit,	1930,	0,	"on") 		
  addCase(1925,	drop,	1925,	0,	"as") 		
  addCase(1925,	emit,	1925,	0,	"<value>") 		# alias

  addCase(1930,	emit,	1930,	0,	"=") 			
  addCase(1930,	call,	1930,	1950,	"(") 		
  addCase(1930,	call,	300,	0,	"select") 		
  addCase(1930,	emit,	1930,	0,	">") 		
  addCase(1930,	skip,	0,	0,	"<key>") 	
  addCase(1930,	emit,	1940,	0,	"<value>") 	

  addCase(1940,	emit,	1930,	0,	".") 		
  addCase(1940,	emit,	1930,	0,	"=") 		
  addCase(1940,	emit,	1930,	0,	">") 		
  addCase(1940,	emit,	1930,	0,	"<") 		
  addCase(1940,	call,	1930,	1940,	"(") 		
  addCase(1940,	repl,	1930,	0,	"where") 		
  addCase(1940,	emit,	1940,	0,	"not") 		
  addCase(1940,	emit,	1940,	0,	"is") 		
  addCase(1940,	emit,	1940,	0,	"null") 		
  addCase(1940,	emit,	1930,	0,	"like") 		
  addCase(1940,	emit,	1960,	1970,	"in") 		
  addCase(1940,	rtne,	0,	0,	")") 		
  addCase(1940,	emit,	1930,	0,	"and") 	
  addCase(1940,	drop,	1923,	0,	"inner") 	
  addCase(1940,	emit,	1930,	0,	"or") 	
  addCase(1940,	drop,	1940,	UNSP,	"on") 	
  addCase(1940,	drop,	1942,	UNSP,	"left") 		
  addCase(1940,	skip,	0,	0,	"<key>") 	
  addCase(1940,	emit,	1940,	0,	"<value>") 	

  addCase(1942,	drop,	1943,	UNSP,	"outer") 		
  addCase(1943,	drop,	1944,	0,	"join") 	
  addCase(1944,	drop,	1945,	0,	"<value>") 	
  addCase(1945,	isrt,	1940,	0,	"and") 	
  
  addCase(1950,	emit,	1930,	0,	"and") 	
  addCase(1950,	emit,	1950,	0,	"is") 	
  addCase(1950,	emit,	1950,	0,	"null") 	
  addCase(1950,	emit,	1930,	0,	"or") 	
  addCase(1950,	rtne,	0,	0,	")") 		
  addCase(1950,	skip,	0,	0,	"<any>") 	

  addCase(1960,	call,	1960,	1970,	"(") 		
  addCase(1960,	call,	300,	1970,	"select") 		
  addCase(1960,	emit,	1970,	0,	"<value>") 	

  addCase(1970,	emit,	1930,	0,	"and") 		
  addCase(1970,	emit,	1930,	0,	"or") 	
  addCase(1970,	rtne,	0,	0,	")") 		
  addCase(1970,	skip,	0,	0,	"<key>") 	
  addCase(1970,	emit,	1970,	0,	"<value>") 	

# Alter function

  addCase(2000,	elog,	2002,	0,	"<value>") 	
  addCase(2002,	isig,	2003,	0,	"<any>") 	
  addCase(2003,	isrt,	2004,	0,	"; create function ") 	
  addCase(2004,	plog,	242,	0,	"<any>") 	

# Alter column

  addCase(2100,	clog,	2102,	0,	"<value>") 	

  addCase(2102,	emit,	2110,	0,	"type") 	
  addCase(2102,	isrt,	2110,	0,	"type") 	

  addCase(2110,	emit,	2110,	0,	"int") 	
  addCase(2110,	repl,	2110,	0,	"uniqueidentifier") 	
  addCase(2110,	repl,	2110,	0,	"money") 	
  addCase(2110,	repl,	2110,	0,	"bit") 	
  addCase(2110,	emit,	2110,	0,	"varchar") 	
  addCase(2110,	repl,	2110,	0,	"nvarchar") 	
  addCase(2110,	repl,	2110,	0,	"datetime") 	
  addCase(2110,	call,	1894,	0,	"(") 	
  addCase(2110,	drop,	2120,	0,	"not") 		# not null
  addCase(2110,	drop,	2130,	0,	"null") 	# null
  addCase(2110,	skip,	0,	0,	"<any>")

  addCase(2120,	isrt,	2121,	0,	",alter column")  
  addCase(2121,	cprt,	2122,	0,	"<any>") 	
  addCase(2122,	isrt,	2123,	0,	"set not") 	
  addCase(2123,	emit,	2110,	0,	"null") 	

  addCase(2130,	isrt,	2131,	0,	",alter column")  
  addCase(2131,	cprt,	2132,	0,	"<any>") 	
  addCase(2132,	isrt,	2110,	0,	"drop not null") 	

# Alter trigger

  addCase(2200,	elog,	2201,	0,	"<value>") 	# Trigger
  addCase(2201,	isrt,	2202,	0,	"on") 	
  addCase(2202,	tprt,	2203,	0,	"<any>") 	
  addCase(2203,	isrt,	2204,	0,	";drop rule if exists") 	
  addCase(2204,	eprt,	2205,	0,	"<any>") 	
  addCase(2205,	isrt,	2206,	0,	"on") 	
  addCase(2206,	tprt,	2207,	0,	"<any>") 	
  addCase(2207,	isrt,	2208,	0,	";drop function") 	
  addCase(2208,	eprt,	2209,	0,	"<any>") 	
  addCase(2209,	isig,	2210,	0,	"<any>") 	
  addCase(2210,	isrt,	2211,	0,	";create") 	
  addCase(2211,	jump,	290,	0,	"<any>") 	

# Alter database

  addCase(2300,	emit,	2301,	0,	"<value>") 	# Database
  addCase(2301,	drop,	1100,	0,	"set") 

# Information Schema processing

  addCase(2400,	emit,	2400,	0,	".") 	
  addCase(2400,	emit,	2410,	0,	"<value>") 	

  addCase(2410,	drop,	2400,	UNSP,	"as") 	
  addCase(2410,	emit,	2412,	0,	"where") 	

  addCase(2412,	emit,	2412,	0,	"(") 	
  addCase(2412,	emit,	2414,	0,	"<value>") 	
  addCase(2414,	emit,	2412,	0,	".") 	
  addCase(2414,	emit,	2416,	0,	"=") 	
  addCase(2414,	emit,	2416,	0,	"like") 	
  addCase(2414,	emit,	2416,	0,	"<") 	

  addCase(2416,	emit,	2416,	0,	">") 	
  addCase(2416,	lowr,	2418,	0,	"<value>") 	

  addCase(2418,	emit,	2412,	0,	"and")
  addCase(2418,	emit,	2412,	0,	"or")
  addCase(2418,	rtne,	2412,	0,	")")

#
# Token scanner 
# Character class table 
#
is_table = [
# 	 0 ^a ^b ^c ^d ^e ^f ^g ^h ^i ^j ^k ^l ^m ^n ^o 	
	 9, 0, 0, 0, 0, 0, 0, 0, 0, 3, 3, 0, 0, 3, 0, 0,

# 	^p ^q ^r ^s ^t ^u ^v ^w ^x ^y ^z ^[ ^\ ^] ^^ ^-		
	 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

# 	sp  !  "  #  $  %  &  '  (  )  *  +  ,  -  .  / 	
	 3, 1, 6, 1, 1, 1, 2, 7, 1, 1, 2, 9, 1, 9, 1, 2,

# 	 0  1  2  3  4  5  6  7  8  9  :  ;  <  =  >  ?		
	 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 1, 1, 2, 2, 2, 1,

# 	 @  A  B  C  D  E  F  G  H  I  J  K  L  M  N  O		
	 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,

# 	 P  Q  R  S  T  U  V  W  X  Y  Z  [  \  ]  ^  _		
	 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 1, 1, 1, 2, 5,

# 	 `  a  b  c  d  e  f  g  h  i  j  k  l  m  n  o		
	 0, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,

# 	 p  q  r  s  t  u  v  w  x  y  z  {  |  }  -  .		
	 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 1, 2, 1, 2, 0 ]

in_comment = 0 		# Processing commnet
is_quoted = 0		# Token is quoted
current_token = ''	# Current token 

#
#  Tokenize input line
#
def tokenize(line) :
  global in_comment
  global is_quoted
  global current_token
  doprint = 0
  tokens = []
# Remove unprintable characters
  newline = ""
  for c in line :
    if ord(c) > 127 : c = ' '
    newline = newline + c 
  line = newline
# Remove dbo
  if re.compile(r'dbo.*?\.').search(line) :
    line=re.sub(r'dbo.*?\.','',line)
# Remove != (replaced later)
  if re.compile(r'!=').search(line) :
    line=re.sub(r'!=','<>',line)
# Remove comments
  if re.compile(r'/\*.*\*/').search(line) :
    line=re.sub(r'/\*.*\*/','',line)
  if not re.compile(r'\'.*/\*').search(line) and  is_quoted == 0 :
    if re.compile(r'--').search(line) :
    	if not re.compile(r'\'--\'').search(line) :
	  m = re.compile(r'--').split(line)
	  print "--",m[1]
	  line = m[0]
    if re.compile(r'/\*').search(line) :
	in_comment = in_comment + 1
	doprint = 1
    if re.compile(r'\*/').search(line) :
	in_comment = in_comment - 1
	doprint = 1
    if in_comment or doprint :
	print line
	return None
# Loop thru line
  i = 0
  while i < len(line) :
	c = line[i]		# process double quotes
	i = i + 1
	if is_quoted == 7 and c == "'" and line[i] == "'" :
	    # extra + c added for non dumped strings
	    current_token = current_token + c + c
	    i = i + 1	
	    continue 
	if c == '!' : c = '_'  # replace bangs
	if c == '[' : c = ' '  # replace square brackets
	if c == ']' : c = ' '  # replace square brackets
	if c == '#' : continue # remove number signs
	if c == '@' : c = '_'  # replace at signs
	if c == '/' and ansi_warnings : c = '?'  # replace / zero divide
        if ord(c) > 127 :
	  t = 5 	       # funny character
        else :
	  t = is_table[ord(c)]
# Quote 
	if t == 7 : 
	  # Remove n' construct - not used by psql
	  if current_token.lower() == 'n' : current_token = ''
	  if is_quoted == t :
	    current_token = current_token + c
	    tokens = tokens +  [ current_token ] 
	    current_token = ''
	    is_quoted = 0
	  else :
	  # Fix b' construct - must not be seperated
	    if current_token.lower() == 'b' :
	      current_token = current_token + c
	      is_quoted = t
	      continue
	    if current_token != ''  :
	      tokens = tokens +  [ current_token ] 
	    current_token = c
	    is_quoted = t
	  continue
# Test quoted
	if is_quoted > 0 :
	  current_token = current_token + c
	  continue 
# Imbedded dot
	if c == '.' and re.compile(r'[0-9][0-9]*').match(current_token) :
	  current_token = current_token + c
	  continue
# Delimiting token
	if t == 1 or t == 2 or t == 9: 
	  if current_token != '' :
	    tokens = tokens +  [ current_token ] 
	  tokens = tokens +  [ c ]
	  current_token = '' 
	  continue
# Delimiter 
	if t == 3 : 
	  if current_token != '' :
	    tokens = tokens +  [ current_token ] 
	    current_token = ''
	  continue
# Upper or Lower 
	if t == 4 or t == 5 : 
	  current_token = current_token + c
	  continue
# Number 
	if t == 8 : 
	  current_token = current_token + c
	  continue
# Return tokens
  if current_token  and is_quoted == 0 :
    tokens = tokens +  [ current_token ] 
    current_token = ''
  return tokens
#
#  Process states on a level
#  Return code is as follows:
#  0 = success, 1 = no case found, 2 = retain token  
#
def processLevel(token) :
  global current_level
  global in_declare
  global begin_count
  global level_stack
  global statement_table
  global current_statement
# Process declare block
  if current_level == 0 and in_declare :
    if token != "declare" :
      printToken("begin")
      in_declare = 0 
      begin_count = begin_count + 1
#      printMessage("begin","inserted",0)
  for case in cases :			# check the table
     if case.level == current_level : 	# for this level
       if case.action == isrt or case.action == isrb :	
	 rc=case.action(case.token,case.move,case.alter)
         return rc
       if case.token == "<key>" :		# keyword search
	   if statement_table.has_key(token.lower()) :
	     rc = case.action(token,case.move,case.alter)
	     return rc
       if case.token == "<any>" :		# token not found
	 rc = case.action(token,case.move,case.alter)
	 return rc
       if case.token == "<symbol>" and token[0] == '_' :# symbol found
	 rc = case.action(token,case.move,case.alter)
	 return rc
       if case.token == token.lower() or case.token == "<value>" :
	 if current_level == 0 :		# token match
	   if statement_table.has_key(token.lower()) :
		statement_table[token.lower()] = statement_table[token.lower()] + 1
	   current_statement = token
	 rc = case.action(token,case.move,case.alter)
         return rc
  return 1
#
#  Process a token on a level
#
def processToken(token) :
  global current_level
  global level_stack
  global trace
  global skip_errors
  global total_errors
  global trap_stack
  while 1 :
    ret = processLevel(token) 
    if ret > 0 :		# no matching case found
      if ret == 1 : 		# no skip
        level_stack = []	# reset stack
        current_level = 0	# level 0
	trap_stack = [] 	# Reset trap stack 
	if skip_errors == 0 :
          print "ERROR ++>" , token, "<", current_level, ">"
	  total_errors = total_errors + 1
          skip_errors = 1
	else :
          printToken(token)
        return
    else :
      if trace :
        print "/*<",current_level,">*/", 
      skip_errors = 0
      return
#
#  Process input file
#
def processFile(filename) :
  global debug
  global current_line
  directory = "."
  filepath = os.path.join(directory , filename)
  for line in file(filepath, 'rU') :
# Print line numbers for debugging
    if trace : print "--",current_line,"]"
    current_line = current_line + 1
    tokens = tokenize(line)
    if tokens :
      for token in  tokens :
        processToken(token)
# Print newline at end of line 
    if is_quoted == 0 : print
  return
#
# Read the int return list
#
def readIntRets() :
  directory = "."
  filepath = os.path.join(directory , "retlist.db")
  for line in file(filepath, 'rU') :
      intret_table[line.rstrip("\n")] = 0
#
# Read the function database
#
def readFuncDB() :
  directory = "."
  filepath = os.path.join(directory , "funclist.db")
  for line in file(filepath, 'rU') :
    line = line.rstrip("\n")
    i = string.find(line,"(")
    proc_table[line[0:i]] = line[i:]
#    print "/*",line[0:i],":",line[i:],"*/"
#
# Read the trigger database
#
def readTrigDB() :
  directory = "."
  filepath = os.path.join(directory , "triglist.db")
  for line in file(filepath, 'rU') :
    line = line.rstrip("\n")
    i = string.find(line,"|")
    trig_table[line[0:i]] = line[i+1:]
#    print "/*",line[0:i],":",line[i+1:],"*/"
#
# Read the table database
#
def readTabDB() :
  directory = "."
  filepath = os.path.join(directory , "tablist.db")
  for line in file(filepath, 'rU') :
    line = line.rstrip("\n")
    i = string.find(line,"|")
    tab_table[line[0:i]] = line[i+1:]
#    print "/*",line[0:i],":",line[i+1:],"*/"

#
# Main program
#
def main() :
  global debug
  global trace
  global total_errors
  global statement_table
  filename = None
  try :
    filename = sys.argv[1]
  except IndexError :
    print "Usage: %s <FILENAME> | build [ trace | debug ]" % sys.argv[0]
    sys.exit(0)
  try :
    if sys.argv[1] == "build" :
  	addAllCases() 
    	printCases()
	return
    if sys.argv[2] == "debug" :
	debug = 1
	trace = 1
    if sys.argv[2] == "trace" :
	trace = 1
  except IndexError :
    debug = 0
    trace = 0
  addAllCases() 
  addAllRepls() 
  readIntRets()
  readFuncDB()
  readTrigDB()
  readTabDB()
  for keycase in cases :
    if keycase.level == 0 :
      statement_table[keycase.token] = 0
  print "create or replace function msload ()"
  print "returns void as $main$ begin"
  print
  processFile(filename)
  print
  print "end; $main$ language plpgsql;"
#
# Print out a block of stats in a comment 
#
  total = 0
  print "-- MSSQL Parser Statement Statistics"
  print
  for key in statement_table :
#	print "-- %12s  : %d" % (key,statement_table[key])
	total = total + statement_table[key]
  print
  print "-- Total  Statements     : ",total
  print "-- Total  Errors Recorded: ",total_errors
#
# Write function database
#
  proc_file = open('funclist.db','w')
  for key in proc_table :
    proc_file.write(key)
    proc_file.write(proc_table[key])
    proc_file.write("\n")
#
# Write trigger database
#
  proc_file = open('triglist.db','w')
  for key in trig_table :
    proc_file.write(key)
    proc_file.write("|")
    proc_file.write(trig_table[key])
    proc_file.write("\n")
#
# Write table database
#
  proc_file = open('tablist.db','w')
  for key in tab_table :
    proc_file.write(key)
    proc_file.write("|")
    proc_file.write(tab_table[key])
    proc_file.write("\n")

  sys.exit(total_errors)

main()

